#######################################################################
#                              Timestep                               #
#######################################################################

function compute_maxspeed(mesh, equation, cache)
  @unpack v = get_static_variables(cache)
  broadcast_volume!(maxspeed, equation, mesh)
  vmax = dg1d.absolute_maximum(v)
  vmax_limit = 1e4
  if vmax > vmax_limit
    @warn "Limiting timestep due to maximum speed exceeding $vmax_limit"
    vmax = vmax_limit
  end
  return vmax
end

function timestep(env, P::Project, hrsc::Maybe{HRSC.AbstractHRSC}, mesh::Mesh1d)
  vmax = compute_maxspeed(mesh, P.equation, mesh.cache)
  dl = min_grid_spacing(mesh)
  dt = dl / (vmax * dtfactor(mesh))
  return dt
end
dtfactor(mesh::Mesh1d{FVElement}) = 2
dtfactor(mesh::Mesh1d{SpectralElement}) = mesh.element.N


function timestep(env, P::Project, hrsc::HRSC.AbstractArtificialViscosity, ::Mesh1d)
  @unpack cache, mesh = env
  @unpack equation = P

  broadcast_volume!(maxspeed, equation, mesh)
  v = get_variable(cache, :v)
  vmax = dg1d.absolute_maximum(view(v,:))
  vmax_limit = 1e4
  if vmax > vmax_limit
    @warn "Limiting timestep due to maximum speed exceeding $vmax_limit"
    vmax = vmax_limit
  end
  smoothed_mu = get_variable(cache, :smoothed_mu)
  mumax = dg1d.absolute_maximum(view(smoothed_mu, :))

  @unpack dl, x, element = mesh
  @unpack N = element

  dt = 1 / (vmax * N / dl + mumax * N^3 / dl^2)
  dt_limit = 1e-7
  if dt < dt_limit
    @warn "Limiting timestep due to dt being smaller than $dt_limit"
    dt = dt_limit
  end
  return dt
end


function timestep(env, P::Project2d, hrsc::Maybe{HRSC.AbstractHRSC}, ::Mesh2d)
  @unpack cache, mesh = env
  @unpack equation = P

  broadcast_volume!(maxspeed, equation, mesh)
  v = get_variable(cache, :v)
  vmax = dg1d.absolute_maximum(view(v,:))
  vmax_limit = 1e4
  if vmax > vmax_limit
    @warn "Limiting timestep due to maximum speed exceeding $vmax_limit"
    vmax = vmax_limit
  end

  @unpack element = mesh
  @unpack N, z = element

  dx, dy = dg1d.widths(mesh.boxes[1])
  for box in view(mesh.boxes, 2:length(mesh.boxes))
    ddx, ddy = dg1d.widths(box)
    ddx < dx && (dx = ddx)
    ddy < dy && (dy = ddy)
  end
  dz = z[2]-z[1]
  dl = min(dx, dy) * dz

  dt = dl / (N * vmax)

  return dt
end


#######################################################################
#                           RHS computation                           #
#######################################################################


function rhs!(env, P::Project, hrsc::Nothing, ::Mesh1d, bdryconds, ldg_bdryconds, av_bdryconds)

  @unpack cache, mesh                   = env
  @unpack equation                      = P
  @unpack rho, q, E                     = get_dynamic_variables(cache)
  @unpack flx_rho, flx_q, flx_E, p, eps = get_static_variables(cache)
  @unpack rhs_rho, rhs_q, rhs_E         = get_rhs_variables(cache)
  @unpack nflx_rho, nflx_q , nflx_E     = get_bdry_variables(cache)
  @unpack bdry_rho, bdry_q, bdry_E,
          bdry_v                        = get_bdry_variables(mesh.cache)
  @unpack v                             = get_static_variables(mesh.cache)

  broadcast_volume!(flux, equation, mesh)

  dg1d.interpolate_face_data!(mesh, rho, bdry_rho)
  dg1d.interpolate_face_data!(mesh, q, bdry_q)
  dg1d.interpolate_face_data!(mesh, E, bdry_E)
  dg1d.interpolate_face_data!(mesh, v, bdry_v)
  broadcast_faces!(llf, equation, mesh)
  broadcast_bdry!(bdryllf, equation, mesh)

  compute_rhs_weak_form!(rhs_rho, flx_rho,  nflx_rho, mesh)
  compute_rhs_weak_form!(rhs_q,   flx_q,    nflx_q,   mesh)
  compute_rhs_weak_form!(rhs_E,   flx_E,    nflx_E,   mesh)

  return
end


function rhs!(env, P::Project, hrsc::Nothing, ::Mesh1d{FVElement}, bdryconds, ldg_bdryconds, av_bdryconds)

  @unpack cache, mesh              = env
  @unpack equation                 = P
  @unpack rho, q, E                = get_dynamic_variables(cache)
  @unpack flx_rho, flx_q, flx_E    = get_static_variables(cache)
  @unpack rhs_rho, rhs_q, rhs_E    = get_rhs_variables(cache)
  @unpack nflx_rho, nflx_q, nflx_E = get_bdry_variables(cache)
  @unpack bdry_rho, bdry_q, bdry_E = get_bdry_variables(cache)
  @unpack dt                       = get_global_variables(cache)

  broadcast_volume!(flux, equation, mesh)
  broadcast_bdry!(fv_bdry_flux, equation, mesh)

  fv_update_step!(rhs_rho, rho, flx_rho, bdry_rho, nflx_rho, dt[1], mesh)
  fv_update_step!(rhs_q,   q,   flx_q,   bdry_q,   nflx_q,   dt[1], mesh)
  fv_update_step!(rhs_E,   E,   flx_E,   bdry_E,   nflx_E,   dt[1], mesh)

  return
end



function rhs!(env, P::Project2d, hrsc::Nothing, ::Mesh2d, bdryconds, ldg_bdryconds, av_bdryconds)

  @unpack cache, mesh                        = env
  @unpack equation                           = P
  @unpack rho, qx, qy, E                     = get_dynamic_variables(cache)
  @unpack flx_rho_x, flx_rho_y,
          flx_qx_x, flx_qx_y,
          flx_qy_x, flx_qy_y,
          flx_E_x, flx_E_y, p, eps           = get_static_variables(cache)
  @unpack rhs_rho, rhs_qx, rhs_qy, rhs_E     = get_rhs_variables(cache)
  @unpack nflx_rho, nflx_qx, nflx_qy, nflx_E = get_bdry_variables(cache)

  @unpack bdry_rho, bdry_qx, bdry_qy, bdry_E,
          bdry_v                             = get_bdry_variables(mesh.cache)
  @unpack v                                  = get_static_variables(mesh.cache)

  broadcast_volume!(flux, equation, mesh)

  dg1d.interpolate_face_data!(mesh, rho, bdry_rho)
  dg1d.interpolate_face_data!(mesh, qx, bdry_qx)
  dg1d.interpolate_face_data!(mesh, qy, bdry_qy)
  dg1d.interpolate_face_data!(mesh, E, bdry_E)
  dg1d.interpolate_face_data!(mesh, v, bdry_v)
  broadcast_faces!(llf, equation, mesh)

  compute_rhs_weak_form!(rhs_rho, flx_rho_x, flx_rho_y,  nflx_rho, mesh)
  compute_rhs_weak_form!(rhs_qx,  flx_qx_x,  flx_qx_y,   nflx_qx,  mesh)
  compute_rhs_weak_form!(rhs_qy,  flx_qy_x,  flx_qy_y,   nflx_qy,  mesh)
  compute_rhs_weak_form!(rhs_E,   flx_E_x,   flx_E_y,    nflx_E,   mesh)

  return
end


function rhs!(env, P::Project, hrsc::AbstractReconstruction, ::Mesh1d, bdryconds, ldg_bdryconds, av_bdryconds)

  @unpack cache, mesh                   = env
  @unpack equation                      = P
  @unpack rho, q, E                     = get_dynamic_variables(cache)
  @unpack flx_rho, flx_q, flx_E, p, eps = get_static_variables(cache)
  @unpack flag                          = get_cell_variables(cache)
  @unpack rhs_rho, rhs_q, rhs_E         = get_rhs_variables(cache)
  @unpack nflx_rho, nflx_q, nflx_E      = get_bdry_variables(cache)
  @unpack bdry_rho, bdry_q, bdry_E,
          bdry_v                        = get_bdry_variables(mesh.cache)
  @unpack v                             = get_static_variables(mesh.cache)

  isperiodic(x) = true
  HRSC.reconstruct!(rho,  flag, hrsc, isperiodic=isperiodic(bdryconds))
  HRSC.reconstruct!(q,    flag, hrsc, isperiodic=isperiodic(bdryconds))
  HRSC.reconstruct!(E,    flag, hrsc, isperiodic=isperiodic(bdryconds))

  broadcast_volume!(flux, equation, mesh)

  dg1d.interpolate_face_data!(mesh, rho, bdry_rho)
  dg1d.interpolate_face_data!(mesh, q, bdry_q)
  dg1d.interpolate_face_data!(mesh, E, bdry_E)
  dg1d.interpolate_face_data!(mesh, v, bdry_v)
  broadcast_faces!(llf, equation, mesh)
  broadcast_bdry!(bdryllf, equation, mesh)

  compute_rhs_weak_form!(rhs_rho, flx_rho,  nflx_rho, mesh)
  compute_rhs_weak_form!(rhs_q,   flx_q,    nflx_q,   mesh)
  compute_rhs_weak_form!(rhs_E,   flx_E,    nflx_E,   mesh)

  return
end


function rhs!(env, P::Project, hrsc::HRSC.AbstractArtificialViscosity, ::Mesh1d,
    bdryconds, ldg_bdryconds, av_bdryconds)

  @unpack cache, mesh = env
  @unpack equation, prms = P
  reg = prms.av_regularization

  if reg === :mono
    rhs_mono!(cache, mesh, equation, P)
  elseif reg === :navierstokes
    rhs_navierstokes!(cache, mesh, equation, P)
  elseif reg === :general
    rhs_general!(cache, mesh, equation, P)
  else
    TODO(reg)
  end
end


function rhs_mono!(cache, mesh, equation, P)

  @unpack rho, q, E                = get_dynamic_variables(cache)
  @unpack flx_rho, flx_q, flx_E, p, eps,
          ldg_rho, ldg_q, ldg_E    = get_static_variables(cache)
  @unpack rhs_rho, rhs_q, rhs_E    = get_rhs_variables(cache)
  @unpack nflx_rho, nflx_q, nflx_E = get_bdry_variables(cache)
  @unpack bdry_rho, bdry_q, bdry_E,
          bdry_v,
          bdry_ldg_rho, bdry_ldg_q, bdry_ldg_E,
          bdry_smoothed_mu         = get_bdry_variables(mesh.cache)
  @unpack v, smoothed_mu           = get_static_variables(mesh.cache)

  dg1d.interpolate_face_data!(mesh, rho, bdry_rho)
  dg1d.interpolate_face_data!(mesh, q, bdry_q)
  dg1d.interpolate_face_data!(mesh, E, bdry_E)
  dg1d.interpolate_face_data!(mesh, v, bdry_v)

  ## solve auxiliary equation: q + ∂x u = 0
  broadcast_faces!(mono_ldg_nflux, equation, mesh)
  broadcast_bdry!(mono_ldg_bdryllf, equation, mesh)
  compute_rhs_weak_form!(ldg_rho, rho, nflx_rho, mesh)
  compute_rhs_weak_form!(ldg_q,   q,   nflx_q,   mesh)
  compute_rhs_weak_form!(ldg_E,   E,   nflx_E,   mesh)

  ## compute rhs of regularized equation: ∂t u + ∂x f + ∂x μ q = 0
  broadcast_volume!(flux, equation, mesh)
  broadcast_faces!(llf, equation, mesh)

  dg1d.interpolate_face_data!(mesh, ldg_rho, bdry_ldg_rho)
  dg1d.interpolate_face_data!(mesh, ldg_q, bdry_ldg_q)
  dg1d.interpolate_face_data!(mesh, ldg_E, bdry_ldg_E)
  dg1d.interpolate_face_data!(mesh, smoothed_mu, bdry_smoothed_mu)

  broadcast_volume!(mono_av_flux, equation, mesh)
  broadcast_faces!(mono_av_nflux, equation, mesh)
  broadcast_bdry!(mono_bdryllf, equation, mesh)
  # TODO don't we also need av_bdrylff here?

  compute_rhs_weak_form!(rhs_rho, flx_rho,  nflx_rho, mesh)
  compute_rhs_weak_form!(rhs_q,   flx_q,    nflx_q,   mesh)
  compute_rhs_weak_form!(rhs_E,   flx_E,    nflx_E,   mesh)

  return
end
function rhs_navierstokes!(cache, mesh, equation, P)

  @unpack rho, q, E                        = get_dynamic_variables(cache)
  @unpack flx_rho, flx_q, flx_E, u,
          ldg_u                            = get_static_variables(cache)
  @unpack rhs_rho, rhs_q, rhs_E            = get_rhs_variables(cache)
  @unpack nflx_rho, nflx_q, nflx_E, nflx_u = get_bdry_variables(cache)
  @unpack bdry_rho, bdry_q, bdry_E,
          bdry_v, bdry_u,
          bdry_ldg_u,
          bdry_smoothed_mu                 = get_bdry_variables(mesh.cache)
  @unpack v, smoothed_mu                   = get_static_variables(mesh.cache)

  broadcast_volume!(navierstokes_u, equation, mesh)
  dg1d.interpolate_face_data!(mesh, rho, bdry_rho)
  dg1d.interpolate_face_data!(mesh, q,   bdry_q)
  dg1d.interpolate_face_data!(mesh, E,   bdry_E)
  dg1d.interpolate_face_data!(mesh, v,   bdry_v)
  dg1d.interpolate_face_data!(mesh, u,   bdry_u)

  ## solve auxiliary equation: q + ∂x u = 0
  broadcast_faces!(navierstokes_ldg_nflux, equation, mesh)
  broadcast_bdry!(navierstokes_ldg_bdryllf, equation, mesh)
  compute_rhs_weak_form!(ldg_u, u, nflx_u, mesh)

  ## compute rhs of regularized equation: ∂t u + ∂x f + ∂x μ q = 0
  broadcast_volume!(flux, equation, mesh)
  broadcast_faces!(llf, equation, mesh)

  dg1d.interpolate_face_data!(mesh, ldg_u,       bdry_ldg_u)
  dg1d.interpolate_face_data!(mesh, smoothed_mu, bdry_smoothed_mu)

  broadcast_volume!(navierstokes_av_flux, equation, mesh)
  broadcast_faces!(navierstokes_av_nflux, equation, mesh)
  broadcast_bdry!(bdryllf, equation, mesh)
  # TODO don't we also need av_bdrylff here?

  compute_rhs_weak_form!(rhs_rho, flx_rho,  nflx_rho, mesh)
  compute_rhs_weak_form!(rhs_q,   flx_q,    nflx_q,   mesh)
  compute_rhs_weak_form!(rhs_E,   flx_E,    nflx_E,   mesh)

  return
end
function rhs_general!(cache, mesh, equation)

  TODO()

  @unpack rho, q, E                            = get_dynamic_variables(cache)
  @unpack flx_rho, flx_q, flx_E, p, eps,
          ldg_rho, ldg_rhoe, ldg_u,
          flx_ldg_rho, flx_ldg_q, flx_ldg_E    = get_static_variables(cache)
  @unpack rhs_rho, rhs_q, rhs_E                = get_rhs_variables(cache)
  @unpack nflx_rho, nflx_q, nflx_E,
          nflx_ldg_rho, nflx_ldg_q, nflx_ldg_E = get_bdry_variables(cache)
  @unpack bdry_rho, bdry_q, bdry_E,
          bdry_v,
          bdry_ldg_rho, bdry_ldg_rhoe, bdry_ldg_u,
          bdry_smoothed_mu                     = get_bdry_variables(mesh.cache)
  @unpack v, smoothed_mu                       = get_static_variables(mesh.cache)

  broadcast_volume!(mono_rhoeps, equation, mesh)
  dg1d.interpolate_face_data!(mesh, rho,    bdry_rho)
  dg1d.interpolate_face_data!(mesh, rhoeps, bdry_rhoeps)
  dg1d.interpolate_face_data!(mesh, u,      bdry_u)
  dg1d.interpolate_face_data!(mesh, v,      bdry_v)

  ## solve auxiliary equation: q + ∂x u = 0
  broadcast_faces!(general_ldg_nflux, equation, mesh)
  broadcast_bdry!(general_ldg_bdryllf, equation, mesh)
  compute_rhs_weak_form!(ldg_rho, rho,    nflx_rho,    mesh)
  compute_rhs_weak_form!(ldg_q,   rhoeps, nflx_rhoeps, mesh)
  compute_rhs_weak_form!(ldg_E,   u,      nflx_u,      mesh)

  ## compute rhs of regularized equation: ∂t u + ∂x f + ∂x μ q = 0
  broadcast_volume!(flux, equation, mesh)
  broadcast_faces!(llf, equation, mesh)

  dg1d.interpolate_face_data!(mesh, ldg_rho,    bdry_ldg_rho)
  dg1d.interpolate_face_data!(mesh, ldg_rhoeps, bdry_ldg_rhoeps)
  dg1d.interpolate_face_data!(mesh, ldg_u,      bdry_ldg_u)
  dg1d.interpolate_face_data!(mesh, smoothed_mu, bdry_smoothed_mu)

  broadcast_volume!(general_av_flux, equation, mesh)
  broadcast_faces!(general_av_nflux, equation, mesh)
  broadcast_bdry!(general_bdryllf, equation, mesh)
  # TODO don't we also need av_bdrylff here?

  compute_rhs_weak_form!(rhs_rho, flx_rho,  nflx_rho, mesh)
  compute_rhs_weak_form!(rhs_q,   flx_q,    nflx_q,   mesh)
  compute_rhs_weak_form!(rhs_E,   flx_E,    nflx_E,   mesh)

  return
end


function rhs!(env, P::Project2d, hrsc::HRSC.AbstractArtificialViscosity, mesh::Mesh2d,
    bdryconds, ldg_bdryconds, av_bdryconds)

  @unpack equation                       = P
  @unpack rho, qx, qy, E                 = get_dynamic_variables(mesh.cache)
  @unpack rhs_rho, rhs_qx, rhs_qy, rhs_E = get_rhs_variables(mesh.cache)
  @unpack flx_rho_x, flx_rho_y,
          flx_qx_x, flx_qx_y,
          flx_qy_x, flx_qy_y,
          flx_E_x, flx_E_y,
          ldg_rho_x, ldg_rho_y,
          ldg_qx_x, ldg_qx_y,
          ldg_qy_x, ldg_qy_y,
          ldg_E_x, ldg_E_y               = get_static_variables(mesh.cache)
  @unpack nflx_rho, nflx_qx, nflx_qy, nflx_E,
          nflx_rho_x, nflx_rho_y,
          nflx_qx_x, nflx_qx_y,
          nflx_qy_x, nflx_qy_y,
          nflx_E_x, nflx_E_y             = get_bdry_variables(mesh.cache)

  @unpack bdry_rho, bdry_qx, bdry_qy, bdry_E,
          bdry_v,
          bdry_ldg_rho_x, bdry_ldg_rho_y,
          bdry_ldg_qx_x, bdry_ldg_qx_y,
          bdry_ldg_qy_x, bdry_ldg_qy_y,
          bdry_ldg_E_x, bdry_ldg_E_y,
          bdry_smoothed_mu               = get_bdry_variables(mesh.cache)
  @unpack v, smoothed_mu                 = get_static_variables(mesh.cache)

  ## TODO Maybe implement a macro with this syntax
  # @weakform(mesh) ∂t ldg_rho_x + ∂x rho = [ nflx_rho_x ]
  # -> compute_rhs_weak_form(ldg_rho_x, rho, 0, nflx_rho_x, mesh)
  # @weakform(mesh) ∂t ldg_rho_y + ∂y rho = [ nflx_rho_y ]
  # @weakform(mesh) ∂t ldg_qx_x  + ∂x qx  = [ nflx_qx_x ]
  # @weakform(mesh) ∂t ldg_qx_y  + ∂y qx  = [ nflx_qx_y ]
  # @weakform(mesh) ∂t ldg_qy_x  + ∂x qy  = [ nflx_qy_x ]
  # @weakform(mesh) ∂t ldg_qy_y  + ∂y qy  = [ nflx_qy_y ]
  # @weakform(mesh) ∂t ldg_E_x   + ∂x E   = [ nflx_E_x ]
  # @weakform(mesh) ∂t ldg_E_y   + ∂y E   = [ nflx_E_y ]

  dg1d.interpolate_face_data!(mesh, rho, bdry_rho)
  dg1d.interpolate_face_data!(mesh, qx, bdry_qx)
  dg1d.interpolate_face_data!(mesh, qy, bdry_qy)
  dg1d.interpolate_face_data!(mesh, E, bdry_E)
  dg1d.interpolate_face_data!(mesh, v, bdry_v)

  ## solve auxiliary equation: q + ∂x u = 0
  broadcast_faces!(ldg_nflux, equation, mesh)

  # TODO Need to revise Bdry conditions

  compute_rhs_weak_form!(ldg_rho_x, rho, 0,   nflx_rho_x, mesh)
  compute_rhs_weak_form!(ldg_rho_y, 0,   rho, nflx_rho_y, mesh)
  compute_rhs_weak_form!(ldg_qx_x,  qx,  0,   nflx_qx_x,  mesh)
  compute_rhs_weak_form!(ldg_qx_y,  0,   qx,  nflx_qx_y,  mesh)
  compute_rhs_weak_form!(ldg_qy_x,  qy,  0,   nflx_qy_x,  mesh)
  compute_rhs_weak_form!(ldg_qy_y,  0,   qy,  nflx_qy_y,  mesh)
  compute_rhs_weak_form!(ldg_E_x,   E,   0,   nflx_E_x,   mesh)
  compute_rhs_weak_form!(ldg_E_y,   0,   E,   nflx_E_y,   mesh)

  ## compute rhs of regularized equation: ∂t u + ∂x f + ∂x μ q = 0
  broadcast_volume!(flux, equation, mesh)
  broadcast_faces!(llf, equation, mesh)

  # TODO Need to revise Bdry conditions

  broadcast_volume!(av_flux_2d, equation, mesh)

  dg1d.interpolate_face_data!(mesh, ldg_rho_x, bdry_ldg_rho_x)
  dg1d.interpolate_face_data!(mesh, ldg_rho_y, bdry_ldg_rho_y)
  dg1d.interpolate_face_data!(mesh, ldg_qx_x, bdry_ldg_qx_x)
  dg1d.interpolate_face_data!(mesh, ldg_qx_y, bdry_ldg_qx_y)
  dg1d.interpolate_face_data!(mesh, ldg_qy_x, bdry_ldg_qy_x)
  dg1d.interpolate_face_data!(mesh, ldg_qy_y, bdry_ldg_qy_y)
  dg1d.interpolate_face_data!(mesh, ldg_E_x, bdry_ldg_E_x)
  dg1d.interpolate_face_data!(mesh, ldg_E_y, bdry_ldg_E_y)
  dg1d.interpolate_face_data!(mesh, smoothed_mu, bdry_smoothed_mu)
  broadcast_faces!(av_nflux_2d, equation, mesh)

  # TODO don't we also need av_bdrylff here?

  compute_rhs_weak_form!(rhs_rho, flx_rho_x, flx_rho_y, nflx_rho, mesh)
  compute_rhs_weak_form!(rhs_qx,  flx_qx_x,  flx_qx_y,  nflx_qx,  mesh)
  compute_rhs_weak_form!(rhs_qy,  flx_qy_x,  flx_qy_y,  nflx_qy,  mesh)
  compute_rhs_weak_form!(rhs_E,   flx_E_x,   flx_E_y,   nflx_E,   mesh)

  return
end
