Project(env::Environment, prms) = Project(env, env.mesh, prms)


function Project(env::Environment, mesh::Mesh1d, prms)

  # construct Project
  eos = EquationOfState.make_EquationOfState(prms["EquationOfState"])
  equation = EulerEquation(eos)
  tci  = TCI.make_TCI(env.mesh, prms["TCI"])
  hrsc = HRSC.make_HRSC(env.mesh, prms["HRSC"])

  bdrycond = dg1d.DirichletBC2()
  fixedprms = (; av_regularization=Symbol(prms["EulerEq"]["av_regularization"]),
                 bernstein=BernsteinReconstruction(mesh))
  P = Project(equation, hrsc, tci, bdrycond, fixedprms)

  # register variables
  # TODO add a ::Nothing overload for register_variables!
  # TODO Somehow replace _register_variables! with register_variables!
  @unpack cache = env
  _register_variables!(mesh, P)
  _register_variables!(mesh, bdrycond)
  _register_variables!(mesh, P.hrsc, prms["EulerEq"]["av_regularization"])
  _register_variables!(mesh, P.tci)
  display(cache)

  # setup initial data
  initialdata!(env, P, prms["EulerEq"])
  # TODO Need to setup some initial viscosity in order to have a chance to even
  # start evolving with the EV method and higher orders

  # setup boundary conditions
  bdryconds, ldg_bdryconds, av_bdryconds = nothing, nothing, nothing

  # TODO Bdry conditions should be moved to project and should only be just a 'lightweight'
  # descriptor. Furthermore, they should also register variables in cache themselves.
  # Note: Atm, only the 'from_initial_data' option would require a place to store the data
  # somewhere. For periodic BCs we don't need extra storage at all.
  # But what other types of boundary conditions will be needed in the future?

  # setup callbacks
  projectcb = make_callback(env, P, bdryconds)
  append!(env.callbacks, CallbackSet(projectcb.callbacks...))

  return P, (bdryconds, ldg_bdryconds, av_bdryconds)
end


function Project(env::Environment, mesh::Mesh2d, prms)

  # construct Project
  eos = EquationOfState.make_EquationOfState(prms["EquationOfState"])
  hrsc = HRSC.make_HRSC(mesh, prms["HRSC"])
  equation = EulerEquation2d(eos)

  bdrycond = dg1d.DirichletBC2()
  P = Project2d(equation, hrsc, nothing)

  # register variables
  # TODO Somehow replace _register_variables! with register_variables!
  @unpack cache = env
  _register_variables!(mesh, P, mesh)
  _register_variables!(mesh, bdrycond)
  _register_variables!(mesh, hrsc, :mono)
  display(cache)

  # setup initial data
  initialdata!(env, P, prms["EulerEq"])

  # setup boundary conditions
  bdryconds_x = nothing
  bdryconds_y = nothing
  ldg_bdryconds, av_bdryconds = nothing, nothing

  # setup callbacks
  projectcb = make_callback(env, P)
  append!(env.callbacks, CallbackSet(projectcb.callbacks...))

  return P, ((bdryconds_x, bdryconds_y), ldg_bdryconds, av_bdryconds)
end


#######################################################################
#                         Register variables                          #
#######################################################################


# Julia limitation/bug: Overloading imported functions without their Module prefix shadows their
# definition in this module without a warning or error. This does not happen for methods
# defined in Base.
# Because of that we prefix with _ below.
# Cf. https://github.com/JuliaLang/julia/issues/15510
function _register_variables!(mesh, P::Project)
  register_variables!(mesh.cache,
    dynamic_variablenames = (:rho, :q, :E,),     # density, momentum density, energy
    rhs_variablenames     = (:rhs_rho, :rhs_q, :rhs_E),
    static_variablenames  = (:flx_rho, :flx_q, :flx_E,
                             :p, :eps, :v,        # pressure, internal energy, speed
                             :S, :Sm1,           # entropy
                             :flx_S, :flx_Sm1),  # entropy flux
    cell_variablenames    = (:max_v,),
    bdry_variablenames    = (:nflx_rho, :nflx_q, :nflx_E,
                             :bdry_rho, :bdry_q, :bdry_E, :bdry_v),
    global_variablenames  = (:t, :tm1) # recent time stamps
  )
end


_register_variables!(mesh, tci_or_hrsc::Nothing, args...) = nothing

_register_variables!(mesh, hrsc::AbstractReconstruction, regularization) = nothing


function _register_variables!(mesh, hrsc::HRSC.AbstractArtificialViscosity, regularization)
  register_variables!(mesh.cache,
                      cell_variablenames    = (:mu,) # 'one viscosity to rule them all'
  )
  if regularization == "mono"
    register_variables!(mesh.cache,
                        static_variablenames  = (:ldg_rho, :ldg_q, :ldg_E,),
                        bdry_variablenames    = (:nflx_ldg_rho, :nflx_ldg_q, :nflx_ldg_E,
                                                 :bdry_ldg_rho, :bdry_ldg_q, :bdry_ldg_E),
    )
  elseif regularization == "navierstokes"
    register_variables!(mesh.cache,
                        static_variablenames  = (:ldg_u, :u),
                        bdry_variablenames = (:bdry_ldg_u, :bdry_u, :nflx_u),
    )
  elseif regularization == "general"
    TODO()
    register_variables!(mesh.cache,
      static_variablenames  = (:ldg_rho, :ldg_q, :ldg_E,              # local DG variable
                               :flx_ldg_rho, :flx_ldg_q, :flx_ldg_E), # local DG flux
      bdry_variablenames    = (:nflx_ldg_rho, :nflx_ldg_q, :nflx_ldg_E), # local DG flux
    )
  end
end


function _register_variables!(mesh::Mesh, hrsc::HRSC.SmoothedArtificialViscosity, regularization)
  _register_variables!(mesh, hrsc.av, regularization)
  register_variables!(mesh.cache,
    static_variablenames  = (:smoothed_mu,),
    bdry_variablenames    = (:bdry_smoothed_mu,)
  )
end


function _register_variables!(mesh::Mesh2d, hrsc::HRSC.EntropyViscosity, regularization)
  register_variables!(mesh.cache,
    static_variablenames = (:S, :Sm1, :flx_S_x, :flx_S_y, :flxm1_S_x, :flxm1_S_y,
                            :ldg_rho_x, :ldg_rho_y, :ldg_qx_x, :ldg_qx_y,
                            :ldg_qy_x,  :ldg_qy_y,  :ldg_E_x,  :ldg_E_y),
    bdry_variablenames   = (:nflx_rho_x, :nflx_rho_y, :nflx_qx_x, :nflx_qx_y,
                            :nflx_qy_x, :nflx_qy_y, :nflx_E_x, :nflx_E_y,
                            :bdry_ldg_rho_x, :bdry_ldg_rho_y, :bdry_ldg_qx_x, :bdry_ldg_qx_y,
                            :bdry_ldg_qy_x, :bdry_ldg_qy_y, :bdry_ldg_E_x, :bdry_ldg_E_y),
    cell_variablenames   = (:mu,),
    global_variablenames = (:tm1,)
  )
end


function _register_variables!(mesh, tci::TCI.AbstractTCI)
  register_variables!(mesh.cache, cell_variablenames = (:flag,))
end


function _register_variables!(mesh, tci::TCI.EntropyProduction)
  register_variables!(mesh.cache,
                      cell_variablenames = (:flag,),
                      static_variablenames = (:EP,), # entropy production
                      bdry_variablenames   = (:EPjump_lhs, :EPjump_rhs)
                                              # entropy production due to jumps at interfaces
  )
end


function _register_variables!(mesh, P::Project2d, ::Mesh2d)
  register_variables!(mesh.cache,
    dynamic_variablenames = (:rho, :qx, :qy, :E),     # density, momentum density, energy
    rhs_variablenames     = (:rhs_rho, :rhs_qx, :rhs_qy, :rhs_E),
    static_variablenames  = (:flx_rho_x, :flx_rho_y,
                             :flx_qx_x, :flx_qx_y,
                             :flx_qy_x, :flx_qy_y,
                             :flx_E_x, :flx_E_y,
                             :p, :eps, :v),        # pressure, internal energy, speed
    bdry_variablenames    = (:nflx_rho, :nflx_qx, :nflx_qy, :nflx_E,
                             :bdry_rho, :bdry_qx, :bdry_qy, :bdry_E, :bdry_v),
    cell_variablenames    = (:max_v,),
    global_variablenames  = (:t,) # recent time stamps
  )
end


# TODO Already overwritten above
# function _register_variables!(cache, bdryconds::Nothing) end
function _register_variables!(mesh::Mesh1d, bdryconds::dg1d.AbstractBC)
  register_variables!(mesh.cache,
    static_variablenames  = (:init_rho, :init_q, :init_E, :init_v,
                             :init_flx_rho, :init_flx_q, :init_flx_E)
  )
end
function _register_variables!(mesh::Mesh2d, bdryconds::dg1d.AbstractBC)
  register_variables!(mesh.cache,
        static_variablenames  = (:init_rho, :init_qx, :init_qy, :init_E, :init_v)
  )
end
