function make_callback(env, P::Project, isperiodic::Bool)
  cbfn_equation(u, t) = callback_equation(u, t, env, P, isperiodic, P.equation)
  cb_equation         = FunctionCallback(cbfn_equation,
                                         CallbackTiming(every_iteration=1,every_dt=0))
  cbfn_tci(u, t)      = callback_tci(u, t, env, P, isperiodic, P.tci)
  cb_tci              = FunctionCallback(cbfn_tci,
                                         CallbackTiming(every_iteration=1, every_dt=0))
  cbfn_hrsc(u, t)     = callback_hrsc(u, t, env, P, isperiodic, P.hrsc)
  cb_hrsc             = FunctionCallback(cbfn_hrsc,
                                         CallbackTiming(every_iteration=1,every_dt=0))
  callbackset = CallbackSet(cb_equation, cb_tci, cb_hrsc)
  return callbackset
end


#######################################################################
#                              Equation                               #
#######################################################################


function callback_equation(state_u, state_t, env, P, isperiodic, eq)
  @unpack mesh        = env
  wrap_dynamic_variables!(mesh, state_u)
  @unpack E, Em1, flr_E, flr_Em1,
          EP, invdetJ = get_static_variables(mesh)
  @unpack t, tm1      = get_global_variables(mesh)
  dt = t[1] - tm1[1]
  tm1 .= t
  t   .= state_t
  Em1     .= E
  flr_Em1 .= flr_E
  broadcast_volume!(entropy_variables, eq, mesh)
  broadcast_volume!(maxspeed, eq, mesh)
  dt <= 0.0 && return

  LL      = layout(mesh)
  EP      = dg1d.vreshape(EP, LL)
  E       = dg1d.vreshape(E, LL)
  flr_E   = dg1d.vreshape(flr_E, LL)
  Em1     = dg1d.vreshape(Em1, LL)
  flr_Em1 = dg1d.vreshape(flr_Em1, LL)
  invdetJ = dg1d.vreshape(invdetJ, LL)

  # compute entropy residual
  @turbo @. flr_Em1 += flr_E
  mul!(EP, mesh.element.D, flr_Em1, 0.5, false)
  @turbo begin
    @. EP .*= invdetJ
    @. EP += (E - Em1) / dt
    @. EP = abs(EP)
  end
end


#######################################################################
#                                HRSC                                 #
#######################################################################


callback_hrsc(state_u, state_t, env, P, isperiodic, hrsc::Nothing) = nothing
callback_hrsc(state_u, state_t, env, P, isperiodic, hrsc::AbstractReconstruction) = nothing


function callback_hrsc(state_u, state_t, env, P, isperiodic, hrsc::Maybe{HRSC.AbstractHRSC})
  TODO(callback_hrsc)
end


function callback_hrsc(state_u, state_t, env, P, isperiodic, hrsc::HRSC.AbstractArtificialViscosity)
  @unpack cache, mesh           = env
  @unpack equation              = P
  @unpack max_v                 = get_static_variables(cache)
  @unpack mu, tmp_mu, cellmax_v = get_cell_variables(cache)
  @unpack D, Sr, τ              = get_dynamic_variables(cache)
  # @unpack vr                    = get_static_variables(cache)
  @unpack t, tm1                = get_global_variables(cache)

  # prevent AV recomputation when callbacks are forced on initial data
  # need this for shock tube tests so that we have non-zero viscosity in troublesome places
  # although we start with smoothed initial data
  t[1] <= tm1[1] && return

  for (k,vi) in enumerate(eachcell(mesh, max_v))
    cellmax_v[k] = maximum(vi)
  end

  HRSC.compute_viscosity!(
      mu, tmp_mu,
      (D,Sr,τ), cellmax_v,
      # (vr,), cellmax_v,
      hrsc)
  mu[1] = mu[end] = 0 # ignore boundaries
end


function callback_hrsc(state_u, state_t, env, P, isperiodic, hrsc::HRSC.TCIViscosity)
  @unpack cache, mesh = env
  @unpack mu, flag    = get_cell_variables(cache)
  HRSC.compute_viscosity!(mu, flag, hrsc)
end


function callback_hrsc(state_u, state_t, env, P, isperiodic, hrsc::HRSC.SmoothedArtificialViscosity)
  callback_hrsc(state_u, state_t, env, P, isperiodic, hrsc.av)
  @unpack cache, mesh = env
  @unpack smoothed_mu = get_static_variables(cache)
  @unpack mu          = get_cell_variables(cache)
  hrsc.smoother(smoothed_mu, mu, mesh, false)
end


function callback_hrsc(state_u, state_t, env, P, isperiodic, hrsc::HRSC.EntropyViscosity)
  @unpack cache, mesh   = env
  @unpack E, flr_E, EP, vr = get_static_variables(cache)
  @unpack mu = get_cell_variables(cache)
  HRSC.compute_viscosity!(mu, E, flr_E, EP, vr, hrsc)
  mu[1] = mu[end] = 0 # ignore boundaries
end


#######################################################################
#                                 TCI                                 #
#######################################################################


function callback_tci(state_u, state_t, env, P, isperiodic, tci::Nothing) end


function callback_tci(state_u, state_t, env, P, isperiodic, tci::TCI.Threshold)
  error("Threshold TCI must be combined with an indicator that is supposed to be zero for " *
        "for smooth solutions, but not for shocks. TODO Implement an indicator variable.")
end


function callback_tci(state_u, state_t, env, P, isperiodic, tci::TCI.EntropyProduction)
  @warn "Figure out entropy fluxes for special relativistic euler equation!"
  TODO(callback_tci)
  @unpack cache, mesh                = env
  @unpack E, Em1, flx_E, flx_Em1, EP = get_static_variables(cache)
  @unpack flag                       = get_cell_variables(cache)
  @unpack EPjump_lhs, EPjump_rhs     = get_bdry_variables(cache)
  @unpack t, tm1                     = get_global_variables(cache)
  TCI.compute_indicator!(flag, EP, EPjump_lhs, EPjump_rhs,
                         E, Em1, flx_E, flx_Em1, t[1], tm1[1],
                         tci)
end


function callback_tci(state_u, state_t, env, P, isperiodic,
    tci::Union{TCI.Diffusion, TCI.Minmod, TCI.Threshold,
               TCI.ModalDecayAverage, TCI.ModalDecayHighest})

  TODO()

  @unpack cache, mesh                        = env
  @unpack sD, sSr, stau                      = get_dynamic_variables(cache)
  @unpack flag, sD_flag, sSr_flag, stau_flag = get_cell_variables(cache)
  TCI.compute_indicator!(sD_flag, sD, tci)
  TCI.compute_indicator!(sSr_flag, sSr, tci)
  TCI.compute_indicator!(stau_flag, stau, tci)
  _, K = layout(mesh)
  for k = 1:K
    sDf, sSrf, stauf = sD_flag[k], sSr_flag[k], stau_flag[k]
    flag[k] = max(sDf, sSrf, stauf)
  end
  # 'smooth' flag with neighbouring cells by simple averaging
  tmp_flag = sD_flag # reuse memory
  for k = 1:K
    k_lhs, k_rhs = periodic_index(k-1, K), periodic_index(k+1, K)
    if !isperiodic && k == 1
      tmp_flag[k] = (flag[k] + flag[k_rhs]) / 2
    elseif !isperiodic && k == K
      tmp_flag[k] = (flag[k_lhs] + flag[k]) / 2
    else
      tmp_flag[k] = (flag[k_lhs] + flag[k] + flag[k_rhs]) / 3
    end
  end
  @. flag = tmp_flag
end
