function make_Equation(eos, prms, m::Mesh)
  @unpack atm_factor, atm_threshold_factor = prms
  if m isa Mesh1d
    return Equation(eos, -1.0, atm_factor, atm_threshold_factor)
  elseif m isa Mesh2d
    TODO()
  else
    error("unknown mesh type encountered")
  end
end


#######################################################################
#                             "spherical1d"                           #
#           Spherical symmetry (1d) + spherical coordinates           #
#######################################################################


@with_signature function flux_source_spherical1d(eq::Equation)
  @accepts D, Sr, τ, ϵ, ρ, p, vr
  @accepts α, ∂αr, βru, ∂βrrdu,
           γrr, γθθ, γϕϕ, ∂γrrr, ∂γrθθ, ∂γrϕϕ, ∂γθϕϕ,
           Γrrr, Γrθθ, Γrϕϕ, Γθrθ, Γθϕϕ, Γϕθϕ, Γϕrϕ,
           Krr, Kθθ, Kϕϕ
  @accepts flr_D, flr_Sr, flr_τ

  γuurr = 1/γrr
  γuuθθ = 1/γθθ
  γuuϕϕ = 1/γϕϕ
  Γkrk = Γrrr + Γθrθ + Γϕrϕ

  gtt = -α^2 + γrr*βru*βru
  gtr = γrr*βru
  grr = γrr
  guutt = -1/α^2
  guutr = βru/α^2
  guurr = γuurr - βru*βru/α^2
  guuθθ = γuuθθ
  guuϕϕ = γuuϕϕ

  vru = γuurr*vr
  v2 = vru*vr
  W = 1/sqrt(1-v2)
  ut = W/α
  ur = W*(vru - βru/α)

  h = 1 + ϵ + p/ρ
  Tuutt = ρ*h*ut*ut + p*guutt
  Tuutr = ρ*h*ut*ur + p*guutr
  Tuurr = ρ*h*ur*ur + p*guurr
  Tuuθθ = p*guuθθ
  Tuuϕϕ = p*guuϕϕ
  Tudtr = Tuutt*gtr + Tuutr*grr

  flr_D  = α*D*(vru - βru/α)
  flr_Sr = α*( Sr*(vru - βru/α) + p )
  flr_τ  = α*( τ*(vru - βru/α) + p*vru )

  src_D = - flr_D*Γkrk

  # tmp1 = 1/2*(Tuutt*βru*βru + 2*Tuutr*βru + Tuurr) * (∂γrrr - 2*Γrrr*γrr)
  tmp1 = 0 # because Dj γik = 0 (Levi-Civita connection)
  tmp2 = Tudtr*(∂βrrdu + Γrrr*βru)
  tmp3 = -Tuutt*α*∂αr
  flθ_Sθ = α*p
  flϕ_Sϕ = α*p
  src_Sr = α*(tmp1 + tmp2 + tmp3) - flr_Sr*Γkrk + flr_Sr*Γrrr + flθ_Sθ*Γθrθ + flϕ_Sϕ*Γϕrϕ

  src_τ = Tuutt*(βru*βru*Krr - βru*∂αr) + Tuutr*(2*βru*Krr - ∂αr) + Tuurr*Krr + Tuuθθ*Kθθ + Tuuϕϕ*Kϕϕ
  src_τ *= α
  src_τ += - flr_τ*Γkrk

  @returns src_D, src_Sr, src_τ, flr_D, flr_Sr, flr_τ
end


@with_signature function maxspeed(equation::Equation)
  @accepts vr, ρ, ϵ
  @accepts α, βru, γrr
  @unpack eos = equation
  cs2 = eos(SquareSpeedOfSound, Density, ρ, InternalEnergy, ϵ)
  cs = sqrt(abs(cs2))
  γuurr = 1/γrr
  vru = γuurr * vr
  v2 = vru*vr
  λ0 = α*vru - βru
  W = 1/sqrt(1-v2)
  λp = α/(1-v2)*(vru*(1-cs2)+cs/W*sqrt((1-v2*cs2)*γuurr-vru^2*(1-cs2)))
  λm = α/(1-v2)*(vru*(1-cs2)-cs/W*sqrt((1-v2*cs2)*γuurr-vru^2*(1-cs2)))
  max_v = absmax(λm, λ0, λp)
  @returns max_v
end


@with_signature function llf(equation::Equation)
  @accepts        D, flr_D, Sr, flr_Sr, τ, flr_τ, max_v, p
  @accepts        α, βru, γrr
  @accepts [bdry] bdry_D, bdry_Sr, bdry_τ, bdry_max_v, bdry_ρ, bdry_vr, bdry_ϵ, bdry_p
  @accepts [bdry] nx

  vmax     = absmax(max_v, bdry_max_v)
  γuurr    = 1/γrr
  bdry_vru = γuurr * bdry_vr

  bdry_flr_D  = α * bdry_D * (bdry_vru - βru/α)
  bdry_flr_Sr = α * ( bdry_Sr * (bdry_vru - βru/α) + bdry_p )
  bdry_flr_τ  = α * ( bdry_τ * (bdry_vru - βru/α) + bdry_p * bdry_vru )

  nflr      = nx*flr_D
  bdry_nflr = nx*bdry_flr_D
  nflr_D    = LLF(nflr, bdry_nflr, D, bdry_D, vmax)

  nflr      = nx*flr_Sr
  bdry_nflr = nx*bdry_flr_Sr
  nflr_Sr   = LLF(nflr, bdry_nflr, Sr, bdry_Sr, vmax)

  nflr      = nx*flr_τ
  bdry_nflr = nx*bdry_flr_τ
  nflr_τ    = LLF(nflr, bdry_nflr, τ, bdry_τ, vmax)

  @returns [bdry] nflr_D, nflr_Sr, nflr_τ
end


@with_signature function bdryllf(equation::Equation)
  @accepts        D, flr_D, Sr, flr_Sr, τ, flr_τ, max_v, vr, p
  @accepts        init_D, init_Sr, init_τ, init_vr, init_max_v, init_p
  @accepts        α, βru, γrr
  @accepts [bdry] nx

  vmax     = absmax(max_v, init_max_v)
  γuurr    = 1/γrr
  init_vru = γuurr * init_vr
  vru      = γuurr * vr

  # enforce dirichlet/neumann conditions following stability analysis
  # see dg-notes/tex/evm.pdf
  # Hesthaven, Numerical Methods for Conservation Laws uses same logic

  # Dirichlet conditions
  # ## if block for hand-tweaked outflow condition for bondi accretion at inner radius
  # if nx > 0
    bdry_D   = -D   + 2*init_D
    bdry_Sr  = -Sr  + 2*init_Sr
    bdry_τ   = -τ   + 2*init_τ
    bdry_vru = -vru + 2*init_vru
    bdry_p   = -p   + 2*init_p
  # else
  #   bdry_D   = D
  #   bdry_Sr  = Sr
  #   bdry_τ   = τ
  #   bdry_vru = vru
  #   bdry_p   = p
  # end

  # Neumann conditions
  # bdry_D   = D
  # bdry_Sr   = Sr
  # bdry_τ = τ

  bdry_flr_D  = α * bdry_D * (bdry_vru - βru/α)
  bdry_flr_Sr = α * ( bdry_Sr * (bdry_vru - βru/α) + bdry_p )
  bdry_flr_τ  = α * ( bdry_τ * (bdry_vru - βru/α) + bdry_p * bdry_vru )

  nflr      = nx*flr_D
  bdry_nflr = nx*bdry_flr_D
  nflr_D    = LLF(nflr, bdry_nflr, D, bdry_D, vmax)

  nflr      = nx*flr_Sr
  bdry_nflr = nx*bdry_flr_Sr
  nflr_Sr   = LLF(nflr, bdry_nflr, Sr, bdry_Sr, vmax)

  nflr      = nx*flr_τ
  bdry_nflr = nx*bdry_flr_τ
  nflr_τ    = LLF(nflr, bdry_nflr, τ, bdry_τ, vmax)

  @returns [bdry] nflr_D, nflr_Sr, nflr_τ
end

@with_signature function fv_bdry_flux(equation::Equation)
  @accepts D, Sr, τ, vr, p
  @accepts init_D, init_Sr, init_τ, init_vr, init_p
  @accepts α, βru, γrr
  @accepts [bdry] nx

  γuurr    = 1/γrr
  init_vru = γuurr * init_vr
  vru      = γuurr * vr

  # Dirichlet conditions
  # if nx < 0
    bdry_D   = -D   + 2*init_D
    bdry_Sr  = -Sr  + 2*init_Sr
    bdry_τ   = -τ   + 2*init_τ
    bdry_vru = -vru + 2*init_vru
    bdry_p   = -p   + 2*init_p
  # else
  #   bdry_D   = D
  #   bdry_Sr  = Sr
  #   bdry_τ   = τ
  #   bdry_vru = vru
  #   bdry_p   = p
  # end

  nflr_D  = α * bdry_D * (bdry_vru - βru/α)
  nflr_Sr = α * ( bdry_Sr * (bdry_vru - βru/α) + bdry_p )
  nflr_τ  = α * ( bdry_τ * (bdry_vru - βru/α) + bdry_p * bdry_vru )

  @returns [bdry] bdry_D, bdry_Sr, bdry_τ, nflr_D, nflr_Sr, nflr_τ
end


#######################################################################
#                       "rescaled_spherical1d"                        #
#           Spherical symmetry (1d) + spherical coordinates           #
#######################################################################


@with_signature function flux_source_rescaled_spherical1d(eq::Equation)
  @accepts D, Sr, τ, p, vr
  @accepts A, ∂Ar, B, ∂Br, r

  γuurr = 1/B^2
  vru   = γuurr * vr

  B3 = B^3
  flr_rD  = r*A*B3 * D * vru
  flr_rSr = r*A*B3 * (Sr*vru + p)
  flr_rτ  = r*A*B3 * (τ + p) * vru

  src_rD  = - A*B3 * D * vru
  # src_rSr = -(τ+p)*r*B3*∂Ar + A*B^2*(r*∂Br+B/3)*(Sr*vru+3*p) - 4/3*A*B3*Sr*vru
  # src_rSr = -(τ+p)*r*B3*∂Ar + A*B^2*(r*∂Br+B/3)*(Sr*vru/A^2+3*p) - 4/3*A*B3*Sr*vru
  src_rSr = -(τ+p)*r*B3*∂Ar + A*B^2*(r*∂Br+B/3)*(Sr*vru/A^2+3*p) - 4/3*A*B3*Sr*vru
  # src_rτ  = -r*B*Sr*∂Ar - A*B3*(τ+p)*vru
  src_rτ  = -r*B*Sr*∂Ar/A - A*B3*(τ+p)*vru

  @returns src_rD, src_rSr, src_rτ, flr_rD, flr_rSr, flr_rτ
end

@with_signature function maxspeed_rescaled_spherical1d(equation::Equation)
  @accepts vr, ρ, ϵ
  @accepts B, A
  @unpack eos = equation
  cs2 = eos(SquareSpeedOfSound, Density, ρ, InternalEnergy, ϵ)
  cs = sqrt(abs(cs2))
  γuurr = 1/B^2
  vru = γuurr * vr
  v2 = vru*vr
  λ0 = A*vru
  W = 1/sqrt(1-v2)
  λp = A/(1-v2)*(vru*(1-cs2)+cs/W*sqrt((1-v2*cs2)*γuurr-vru^2*(1-cs2)))
  λm = A/(1-v2)*(vru*(1-cs2)-cs/W*sqrt((1-v2*cs2)*γuurr-vru^2*(1-cs2)))
  max_v = absmax(λm, λ0, λp)
  @returns max_v
end

@with_signature function llf_rescaled_spherical1d(equation::Equation)
  @accepts        D, flr_rD, Sr, flr_rSr, τ, flr_rτ, max_v, p
  @accepts        A, B, r
  @accepts [bdry] bdry_D, bdry_Sr, bdry_τ, bdry_max_v, bdry_vr, bdry_p
  @accepts [bdry] nx

  vmax     = absmax(max_v, bdry_max_v)
  γuurr    = 1/B^2
  bdry_vru = γuurr * bdry_vr
  B3       = B^3

  rD       = r*B3*D
  rSr      = r*B3*Sr
  rτ       = r*B3*τ
  bdry_rD  = r*B3*bdry_D
  bdry_rSr = r*B3*bdry_Sr
  bdry_rτ  = r*B3*bdry_τ

  bdry_flr_rD  = r*A*B3 * bdry_D * bdry_vru
  bdry_flr_rSr = r*A*B3 * ( bdry_Sr * bdry_vru + bdry_p )
  bdry_flr_rτ  = r*A*B3 * ( (bdry_τ + bdry_p) * bdry_vru )

  nflr      = nx*flr_rD
  bdry_nflr = nx*bdry_flr_rD
  nflr_rD   = LLF(nflr, bdry_nflr, rD, bdry_rD, vmax)

  nflr      = nx*flr_rSr
  bdry_nflr = nx*bdry_flr_rSr
  nflr_rSr  = LLF(nflr, bdry_nflr, rSr, bdry_rSr, vmax)

  nflr      = nx*flr_rτ
  bdry_nflr = nx*bdry_flr_rτ
  nflr_rτ   = LLF(nflr, bdry_nflr, rτ, bdry_rτ, vmax)

  @returns [bdry] nflr_rD, nflr_rSr, nflr_rτ
end

@with_signature function bdryllf_rescaled_spherical1d(equation::Equation)
  @accepts        D, flr_rD, Sr, flr_rSr, τ, flr_rτ, max_v, vr, p
  @accepts        init_D, init_Sr, init_τ, init_vr, init_max_v, init_p
  @accepts        A, B, r
  @accepts [bdry] nx

  vmax     = absmax(max_v, init_max_v)
  γuurr    = 1/B^2
  init_vru = γuurr * init_vr
  vru      = γuurr * vr
  B3       = B^3

  # Dirichlet conditions
  bdry_D   = -D   + 2*init_D
  bdry_Sr  = -Sr  + 2*init_Sr
  bdry_τ   = -τ   + 2*init_τ
  bdry_vru = -vru + 2*init_vru
  bdry_p   = -p   + 2*init_p

  # Neumann conditions
  # bdry_D   = D
  # bdry_Sr   = Sr
  # bdry_τ = τ

  rD       = r*B3*D
  rSr      = r*B3*Sr
  rτ       = r*B3*τ
  bdry_rD  = r*B3*bdry_D
  bdry_rSr = r*B3*bdry_Sr
  bdry_rτ  = r*B3*bdry_τ

  bdry_flr_rD  = r*A*B3 * bdry_D * bdry_vru
  bdry_flr_rSr = r*A*B3 * ( bdry_Sr * bdry_vru + bdry_p )
  bdry_flr_rτ  = r*A*B3 * (bdry_τ + bdry_p) * bdry_vru

  nflr      = nx*flr_rD
  bdry_nflr = nx*bdry_flr_rD
  nflr_rD   = LLF(nflr, bdry_nflr, rD, bdry_rD, vmax)

  nflr      = nx*flr_rSr
  bdry_nflr = nx*bdry_flr_rSr
  nflr_rSr  = LLF(nflr, bdry_nflr, rSr, bdry_rSr, vmax)

  nflr      = nx*flr_rτ
  bdry_nflr = nx*bdry_flr_rτ
  nflr_rτ   = LLF(nflr, bdry_nflr, rτ, bdry_rτ, vmax)

  @returns [bdry] nflr_rD, nflr_rSr, nflr_rτ
end


@with_signature function unscale_conservatives(equation::Equation)
  @accepts rD, rSr, rτ
  @accepts r, B
  B3 = B^3
  D  = rD/(r*B3)
  Sr = rSr/(r*B3)
  τ  = rτ/(r*B3)
  @returns D, Sr, τ
end


@with_signature function fv_bdry_flux_rescaled_spherical1d(equation::Equation)
  @accepts D, Sr, τ, vr, p
  @accepts init_D, init_Sr, init_τ, init_vr, init_p
  @accepts A, B, ∂Br, r
  @accepts [bdry] nx

  γuurr    = 1/B^2
  init_vru = γuurr * init_vr
  vru      = γuurr * vr
  B3       = B^3

  # # Dirichlet conditions
  bdry_D   = -D   + 2*init_D
  bdry_Sr  = -Sr  + 2*init_Sr
  bdry_τ   = -τ   + 2*init_τ
  bdry_vru = -vru + 2*init_vru
  bdry_p   = -p   + 2*init_p
  # bdry_D   = init_D
  # bdry_Sr  = init_Sr
  # bdry_τ   = init_τ
  # bdry_vru = init_vru
  # bdry_p   = init_p

  @show init_D, D

  nflr_rD  = r*A*B3* bdry_D * bdry_vru
  nflr_rSr = r*A*B3* (bdry_Sr * bdry_vru + bdry_p)
  nflr_rτ  = r*A*B3* (bdry_τ + bdry_p) * bdry_vru

  # FIXME We are abusing the bdry_D,Sr,τ buffers to store the rescaled bdry values
  bdry_D  *= r*B3
  bdry_Sr *= r*B3
  bdry_τ  *= r*B3

  @returns [bdry] bdry_D, bdry_Sr, bdry_τ, nflr_rD, nflr_rSr, nflr_rτ
end


#######################################################################
#            helpers to resolve coordinate singularities              #
#            by imposing symmetry conditions at r=0                   #
#######################################################################


function impose_symmetry!(P::Project{:rescaled_spherical1d}, mesh::Mesh)

  @unpack cache = mesh
  @unpack D, Sr, τ = get_static_variables(cache)
  L = layout(mesh)
  Npts, _ = L

  vD  = dg1d.vreshape(D, L)
  vSr = dg1d.vreshape(Sr, L)
  vτ  = dg1d.vreshape(τ, L)

  # even variables are interpolated such that derivative vanishes at r=0
  for var in (vD, vτ)
    tmp = 0
    for j = 2:Npts
      tmp += mesh.element.D[1,j] * var[j,1]
    end
    var[1,1] = -tmp/mesh.element.D[1,1]
  end
  # odd variables are set to zero at r=0
  Sr[1] = 0

  return
end


function impose_symmetry!(P::Project{:rescaled_spherical1d}, mesh::Mesh{FVElement})

  @unpack cache = mesh
  @unpack rD = get_dynamic_variables(cache)
  @unpack D, Sr, τ, r = get_static_variables(cache)
  L = layout(mesh)
  Npts, _ = L

  @show rD[1], rD[2]

  # symmetric variables are interpolated such that derivative vanishes at r=0
  for var in (D, τ)
    tmp = 0
    # use quadratic interpolation around origin
    var2, var3, r2, r3 = var[2], var[3], r[2], r[3]
    var[1] = (r3^2*var2-r2^2*var3)/(r3^2-r2^2)
  end
  # asymmetric variables are set to zero at r=0
  Sr[1] = 0

  @show D[1], D[2]

  return
end


function impose_symmetry_sources!(P::Project{:spherical1d}, mesh::Mesh)
  @unpack r = get_static_variables(mesh.cache)
  r[1] ≈ 0.0 || return
  @unpack src_D, src_Sr, src_τ = get_static_variables(mesh.cache)
  L = layout(mesh)
  Npts, _ = L
  v_src_D  = dg1d.vreshape(src_D, L)
  v_src_τ  = dg1d.vreshape(src_τ, L)
  # even variables are interpolated such that derivative vanishes at r=0
  for var in (v_src_D, v_src_τ)
    tmp = 0
    for j = 2:Npts
      tmp += mesh.element.D[1,j] * var[j,1]
    end
    var[1,1] = -tmp/mesh.element.D[1,1]
  end
  # odd variables are set to zero at r=0
  src_Sr[1] = 0
  return
end


#######################################################################
#                                 LDG                                 #
#######################################################################


@with_signature function ldg_nflux_spherical1d(eq::Equation)
  @accepts        D, Sr, τ
  @accepts [bdry] bdry_D, bdry_Sr, bdry_τ
  @accepts [bdry] nx
  nflr_D  = 0.5*nx*(D+bdry_D)
  nflr_Sr = 0.5*nx*(Sr+bdry_Sr)
  nflr_τ  = 0.5*nx*(τ+bdry_τ)
  @returns [bdry] nflr_D, nflr_Sr, nflr_τ
end


@with_signature function ldg_bdryllf_spherical1d(eq::Equation)
  @accepts        D, Sr, τ
  @accepts [bdry] nx
  nflr_D  = nx*D
  nflr_Sr = nx*Sr
  nflr_τ  = nx*τ
  @returns [bdry] nflr_D, nflr_Sr, nflr_τ
end


@with_signature function av_flux_covariant_spherical1d(equations::Equation)
  @accepts rhs_D, rhs_Sr, rhs_τ, flr_D, flr_Sr, flr_τ, ldg_D, ldg_Sr, ldg_τ, smoothed_mu
  flr_D  += -smoothed_mu * (rhs_D  - ldg_D)
  flr_Sr += -smoothed_mu * (rhs_Sr - ldg_Sr)
  flr_τ  += -smoothed_mu * (rhs_τ  - ldg_τ)
  @returns flr_D, flr_Sr, flr_τ
end


@with_signature function av_nflux_covariant_spherical1d(eq::Equation)
  @accepts        rhs_D, rhs_Sr, rhs_τ, ldg_D, ldg_Sr, ldg_τ, smoothed_mu, r
  @accepts [bdry] bdry_rhs_D, bdry_rhs_Sr, bdry_rhs_τ, bdry_ldg_D, bdry_ldg_Sr, bdry_ldg_τ, bdry_smoothed_mu
  @accepts [bdry] nflr_D, nflr_Sr, nflr_τ, nx

  nflr       = -nx*(rhs_D - ldg_D)
  bdry_nflr  = -nx*(bdry_rhs_D - bdry_ldg_D)
  nflr_D    += 0.5 * (smoothed_mu*nflr + bdry_smoothed_mu*bdry_nflr)

  nflr       = -nx*(rhs_Sr - ldg_Sr)
  bdry_nflr  = -nx*(bdry_rhs_Sr - bdry_ldg_Sr)
  nflr_Sr    += 0.5 * (smoothed_mu*nflr + bdry_smoothed_mu*bdry_nflr)

  nflr       = -nx*(rhs_τ - ldg_τ)
  bdry_nflr  = -nx*(bdry_rhs_τ - bdry_ldg_τ)
  nflr_τ     += 0.5 * (smoothed_mu*nflr + bdry_smoothed_mu*bdry_nflr)

  @returns [bdry] nflr_D, nflr_Sr, nflr_τ
end


#######################################################################
#                              Analysis                               #
#######################################################################


@with_signature function entropy_variables(equation::Equation)
  @accepts ρ, p, vr
  @accepts γrr
  @unpack eos = equation
  K, Gamma = eos.K, eos.Gamma
  γrruu = 1/γrr
  vru = γrruu * vr
  v2 = vr*vru
  # specific entropy as used in EFL paper
  s = log(p/ρ^Gamma)
  # specific entropy computed from EoS as defined in Laim 1977
  # s = (K * (1 + ρ/p))^(1-Gamma)
  ooW = sqrt(1-v2)
  E = s * ρ / ooW
  flr_E = vru * E
  @returns E, flr_E
end
