function initialdata!(env, P::Project, prms)
  @unpack id = prms
  initialdata_equation!(Val(Symbol(id)), env, P, prms)
  # # initialdata_entropy_variables!(id, P, prms)
  # initialdata_hrsc!(Val(Symbol(id)), env, P, P.hrsc, prms)
  # initialdata_tci!(Val(Symbol(id)), env, P, P.tci, prms)
end


#######################################################################
#                     Initial data gravity sector                     #
#######################################################################


function initialdata_equation!(::Val{:bondi_accretion}, env, P::Project{:spherical1d}, prms)

  @unpack mesh = env
  @unpack equation = P

  id_filename = prms["id_filename"]
  if !isfile(id_filename)
    error("GRHD: Failed to locate bondi_accretion initial data at $id_filename")
  end

  data = h5open(id_filename, "r") do file
    g = file["bondi_accretion"]
    Γ  = read_attribute(g, "Γ")
    K  = read_attribute(g, "K")
    ρc = read_attribute(g, "ρc")
    rc = read_attribute(g, "rc")
    M  = read_attribute(g, "M")
    r  = read(g, "r")
    ρ  = read(g, "ρ")
    p  = read(g, "p")
    ϵ  = read(g, "ϵ")
    vr = read(g, "vr")
    (; Γ, K, ρc, rc, M, r, ρ, ϵ, p, vr)
  end

  @assert data.K ≈ equation.eos.K
  @assert data.Γ ≈ equation.eos.Gamma

  @unpack D, Sr, τ = get_dynamic_variables(mesh.cache)
  @unpack ρ, p, ϵ, vr, r, max_v,
          init_D, init_Sr, init_τ, init_vr, init_p,
          init_max_v = get_static_variables(mesh.cache)
  @unpack α, βru, ∂αr, ∂βrrdu,
          γrr, γθθ, γϕϕ, ∂γrrr, ∂γrθθ, ∂γrϕϕ, ∂γθϕϕ,
          Γrrr, Γrθθ, Γrϕϕ, Γθrθ, Γθϕϕ, Γϕrϕ, Γϕθϕ,
          Krr, Kθθ, Kϕϕ = get_static_variables(mesh.cache)

  p   .= interpolate(r, data.r, data.p)
  ρ   .= interpolate(r, data.r, data.ρ)
  ϵ   .= interpolate(r, data.r, data.ϵ)
  vru  = interpolate(r, data.r, data.vr)

  equation.atm_density = maximum(ρ) * equation.atm_factor
  equation.atm_factor = prms["atm_factor"]
  equation.atm_threshold = prms["atm_threshold_factor"]

  # to avoid problems with 1/sin(θ) etc., this is also done in initialdata/SphericalAccretion.jl
  θ = π/2
  @unpack M = data

  @. α   = sqrt(r/(r+2*M))
  @. βru = 2*M/(r+2*M) # = β^r
  @. ∂αr = M / sqrt(r*(2*M+r)^3)
  @. ∂βrrdu = - 2*M/(2*M+r)^2

  # γ_ij
  @. γrr = 1 + 2*M/r
  @. γθθ = r^2
  @. γϕϕ = r^2 * sin(θ)^2

  # # γijuu = γ^ij
  # γrruu = @. 1/γrr
  # γθθuu = @. 1/γθθ
  # γϕϕuu = @. 1/γϕϕ

  # ∂iγjk := ∂γijk
  @. ∂γrrr = -2*M/r^2
  @. ∂γrθθ = 2*r
  @. ∂γrϕϕ = 2*r*sin(θ)^2
  @. ∂γθϕϕ = r^2 * sin(θ)*cos(θ)

  # Γijk := Γ^i_jk wrt γ_ij
  @. Γrrr = -M/(r*(2*M+r))
  @. Γrθθ = -r^2/(2*M+r)
  @. Γrϕϕ = -r^2*sin(θ)^2/(2*M+r)
  @. Γθrθ = 1/r
  @. Γθϕϕ = -sin(θ)*cos(θ)
  @. Γϕrϕ = 1/r
  @. Γϕθϕ = cos(θ)/sin(θ)

  # Kij := K_ij
  @. Krr = -2*M*(M+r)/sqrt(r^5*(2*M+r))
  @. Kθθ = 2*M*sqrt(r/(2*M+r))
  @. Kϕϕ = 2*M*sqrt(r/(2*M+r))*sin(θ)^2

  # # ∂Kijk = ∂_i K_jk
  # ∂Krrr = @. 2*M*(5*M^2+2*r^2+6*M*r)/sqrt(r^7*(2*M+r)^3)
  # ∂Krθθ = @. 2*M^2/sqrt(r*(2*M+r)^3)
  # ∂Krϕϕ = @. ∂Krθθ * sin(θ)^2

  # # verify trace of K
  # trK_v1 = @. Krr*γrruu + Kθθ*γθθuu + Kϕϕ*γϕϕuu
  # trK_v2 = @. 2*M*α^3/r^2*(1+3*M/r)
  # @show(extrema(trK_v1.-trK_v2))


  # from initialdata/SphericalAccretion.jl
  # also see Papadopoulos and Font 1998, arXiv:gr-qc/9803087, eq (5)
  v2  = @. γrr*vru^2
  W   = @. 1/sqrt(1-v2)  # = αu^t, Lorentz factor
  @. vr = γrr * vru # = γ_rr v^r

  h = @. 1 + ϵ + p/ρ
  @. D  = W * ρ
  @. Sr = W^2 * ρ * h * vr
  @. τ  = W^2 * ρ * h - p - D

  broadcast_volume!(cons2prim, equation, mesh)
  broadcast_volume!(maxspeed, equation, mesh)
  broadcast_volume!(flux_source_spherical1d, equation, mesh)

  @. init_D     = D
  @. init_Sr    = Sr
  @. init_τ     = τ
  @. init_vr    = vr
  @. init_p     = p
  @. init_max_v = max_v

  return
end


function initialdata_equation!(::Val{:tov}, env, P::Project{:spherical1d}, prms)

  filename = prms["id_filename"]
  if !isfile(filename)
    error("GRHD: Failed to locate TOV initial data at id_filename = $filename")
  end

  data = h5open(filename, "r") do file
    g = file["TOV"]
    Γ  = read_attribute(g, "Γ")
    K  = read_attribute(g, "K")
    ρc = read_attribute(g, "ρc")
    M  = read_attribute(g, "M")
    rsch = read(g, "r") # schwarzschild radius coordinate
    riso = read(g, "R") # isotropic radius coordinate
    ρ   = read(g, "ρ")
    p   = read(g, "p")
    ϵ   = read(g, "ϵ")
    ϕ   = read(g, "ϕ")
    m   = read(g, "m")

    # metric potentials: ds^2 = -A^2 dt^2 + B^2 γij dx^i dx^j
    A = @. exp(ϕ)
    B = @. rsch / riso
    # B[1] is nan, because rsch = riso = 0 there
    # use a parabola to interpolate data point at riso = 0
    # this also ensures that B is symmetric arcoss riso = 0
    B2, B3, r2, r3 = B[2], B[3], riso[2], riso[3]
    B[1] = (r3^2*B2-r2^2*B3)/(r3^2-r2^2)

    (; Γ, K, ρc, M, riso, ρ, ϵ, p, A, B)
  end

  @unpack equation = P
  @unpack mesh = env
  @unpack eos = P.equation
  @assert data.K ≈ eos.K
  @assert data.Γ ≈ eos.Gamma

  equation.atm_density = data.ρc * equation.atm_factor
  equation.atm_factor = prms["atm_factor"]
  equation.atm_threshold = prms["atm_threshold_factor"]

  @unpack D, Sr, τ = get_dynamic_variables(mesh.cache)
  @unpack ρ, p, ϵ, vr, r, max_v,
          init_D, init_Sr, init_τ, init_vr, init_p,
          init_max_v = get_static_variables(mesh.cache)
  @unpack α, βru, ∂αr, ∂βrrdu,
          γrr, γθθ, γϕϕ, ∂γrrr, ∂γrθθ, ∂γrϕϕ, ∂γθϕϕ,
          Γrrr, Γrθθ, Γrϕϕ, Γθrθ, Γθϕϕ, Γϕrϕ, Γϕθϕ,
          Krr, Kθθ, Kϕϕ = get_static_variables(mesh.cache)

  # need to split initial data into portion that lies inside and outside of star,
  # because derivatives of the hydro variables are discontinuous at the surface
  # although metric functions are supposed to be better conditioned there,
  # we also split them for the interpolation, because of potential contamination
  # with numerical error from patching
  rmax = data.riso[end]
  idx_exterior = findfirst(ri -> ri > rmax, r) |> something
  @assert idx_exterior > 0
  interior = 1:idx_exterior-1
  exterior = idx_exterior:length(mesh)

  rint = view(r, interior)
  rext = view(r, exterior)
  A, B, ∂Ar, ∂Br = [ similar(r) for _ in 1:4 ]

  ρ[interior] .= interpolate(rint, data.riso, data.ρ)
  ϵ[interior] .= interpolate(rint, data.riso, data.ϵ)
  p[interior] .= interpolate(rint, data.riso, data.p)
  A[interior] .= interpolate(rint, data.riso, data.A)
  B[interior] .= interpolate(rint, data.riso, data.B)

  ρ[exterior] .= equation.atm_density
  p[exterior] .= @. eos(Pressure, Density, ρ[exterior])
  ϵ[exterior] .= @. eos(InternalEnergy, Density, ρ[exterior])
  # metric potentials of Schwarzschild solution in isotropic coordinates:
  #   ds^2 = -A^2 dt^2 + B^2 γij dx^i dx^j
  @unpack M = data
  A[exterior] .= @. (1-M/(2*rext))/(1+M/(2*rext))
  B[exterior] .= @. (1+M/(2*rext))^2

  # BasicInterpolators.jl cannot differentiate ...  so we diff numerically
  ∂Ar .= dg1d.differentiate(A, mesh)
  ∂Br .= dg1d.differentiate(B, mesh)

  θ = π/2

  @. α = A
  @. βru = 0.0
  @. ∂αr = ∂Ar
  @. ∂βrrdu = 0

  @. γrr = B^2
  @. γθθ = B^2*r^2
  @. γϕϕ = B^2*r^2*sin(θ)^2

  @. ∂γrrr = 2*B*∂Br
  @. ∂γrθθ = 2*(r*B^2+B*∂Br*r^2)
  @. ∂γrϕϕ = ∂γrθθ * sin(θ)^2
  @. ∂γθϕϕ = 0.0

  @. Γrrr = ∂Br/B
  @. Γrθθ = (-r^2*B*∂Br-r*B^2)/B^2
  @. Γrϕϕ = Γrθθ*sin(θ)^2
  @. Γθrθ = (r^2*B*∂Br+r*B^2)/(r^2*B^2)
  @. Γθϕϕ = -sin(θ)*cos(θ)
  @. Γϕrϕ = (r^2*B*∂Br+r*B^2)/(r^2*B^2)
  @. Γϕθϕ = cos(θ)/sin(θ)

  @. Krr = 0.0
  @. Kθθ = 0.0
  @. Kϕϕ = 0.0

  @. vr = 0
  W = 1.0
  h = @. 1 + ϵ + p/ρ
  @. D  = W * ρ
  @. Sr = W^2 * ρ * h * vr
  @. τ  = W^2 * ρ * h - p - D

  broadcast_volume!(cons2prim, equation, mesh)
  broadcast_volume!(maxspeed, equation, mesh)
  broadcast_volume!(flux_source_spherical1d, equation, mesh)
  impose_symmetry_sources!(P, mesh)

  @. init_D     = D
  @. init_Sr    = Sr
  @. init_τ     = τ
  @. init_vr    = vr
  @. init_p     = p
  @. init_max_v = max_v

end


function initialdata_equation!(::Val{:tov}, env, P::Project{:rescaled_spherical1d}, prms)

  filename = prms["id_filename"]
  if !isfile(filename)
    error("GRHD: Failed to locate TOV initial data at id_filename = $filename")
  end

  data = h5open(filename, "r") do file
    g = file["TOV"]
    Γ  = read_attribute(g, "Γ")
    K  = read_attribute(g, "K")
    ρc = read_attribute(g, "ρc")
    M  = read_attribute(g, "M")
    rsch = read(g, "r") # schwarzschild radius coordinate
    riso = read(g, "R") # isotropic radius coordinate
    ρ   = read(g, "ρ")
    p   = read(g, "p")
    ϵ   = read(g, "ϵ")
    ϕ   = read(g, "ϕ")
    m   = read(g, "m")

    # metric potentials: ds^2 = -A^2 dt^2 + B^2 γij dx^i dx^j
    A = @. exp(ϕ)
    B = @. rsch / riso
    # B[1] is nan, because rsch = riso = 0 there
    # use a parabola to interpolate data point at riso = 0
    # this also ensures that B is symmetric arcoss riso = 0
    B2, B3, r2, r3 = B[2], B[3], riso[2], riso[3]
    B[1] = (r3^2*B2-r2^2*B3)/(r3^2-r2^2)

    (; Γ, K, ρc, M, riso, ρ, ϵ, p, A, B)
  end

  @unpack equation = P
  @unpack mesh = env
  @unpack eos = P.equation
  @assert data.K ≈ eos.K
  @assert data.Γ ≈ eos.Gamma

  equation.atm_density = data.ρc * equation.atm_factor
  equation.atm_factor = prms["atm_factor"]
  equation.atm_threshold = prms["atm_threshold_factor"]

  @unpack rD, rSr, rτ = get_dynamic_variables(mesh.cache)
  @unpack ρ, p, ϵ, vr, r, max_v,
          D, Sr, τ,
          init_D, init_Sr, init_τ,
          init_vr, init_p, init_max_v = get_static_variables(mesh.cache)
  @unpack A, ∂Ar, B, ∂Br = get_static_variables(mesh.cache)

  # need to split initial data into portion that lies inside and outside of star,
  # because derivatives of the hydro variables are discontinuous at the surface
  # although metric functions are supposed to be better conditioned there,
  # we also split them for the interpolation, because of potential contamination
  # with numerical error from patching
  rmax = data.riso[end]
  idx_exterior = findfirst(ri -> ri > rmax, r) |> something
  @assert idx_exterior > 0
  interior = 1:idx_exterior-1
  exterior = idx_exterior:length(mesh)

  rint = view(r, interior)
  rext = view(r, exterior)

  ρ[interior] .= interpolate(rint, data.riso, data.ρ)
  ϵ[interior] .= interpolate(rint, data.riso, data.ϵ)
  p[interior] .= interpolate(rint, data.riso, data.p)
  A[interior] .= interpolate(rint, data.riso, data.A)
  B[interior] .= interpolate(rint, data.riso, data.B)

  ρ[exterior] .= equation.atm_density
  p[exterior] .= @. eos(Pressure, Density, ρ[exterior])
  ϵ[exterior] .= @. eos(InternalEnergy, Density, ρ[exterior])
  # metric potentials of Schwarzschild solution in isotropic coordinates:
  #   ds^2 = -A^2 dt^2 + B^2 γij dx^i dx^j
  @unpack M = data
  A[exterior] .= @. (1-M/(2*rext))/(1+M/(2*rext))
  B[exterior] .= @. (1+M/(2*rext))^2

  # BasicInterpolators.jl cannot differentiate ...  so we diff numerically
  ∂Ar .= dg1d.differentiate(A, mesh)
  ∂Br .= dg1d.differentiate(B, mesh)

  @. vr = 0
  W = 1.0
  h = @. 1 + ϵ + p/ρ
  @. D  = W * ρ
  @. Sr = W^2 * ρ * h * vr
  @. τ  = W^2 * ρ * h - p - D
  @. rD  = r*B^3*D
  @. rSr = r*B^3*Sr
  @. rτ  = r*B^3*τ

  broadcast_volume!(unscale_conservatives, equation, mesh)
  impose_symmetry!(P, mesh)
  broadcast_volume!(cons2prim_rescaled_spherical1d, equation, mesh)
  broadcast_volume!(flux_source_rescaled_spherical1d, equation, mesh)
  broadcast_volume!(maxspeed_rescaled_spherical1d, equation, mesh)

  @. init_D     = D
  @. init_Sr    = Sr
  @. init_τ     = τ
  @. init_vr    = vr
  @. init_p     = p
  @. init_max_v = max_v

end


function interpolate(x, x_ref, y_ref)
  intrp = CubicSplineInterpolator(x_ref, y_ref)
  return intrp.(x)
end


function interpolate_derivative(x, x_ref, y_ref)
  intrp = CubicSplineInterpolator(x_ref, y_ref)
  return intrp.(x)
end


#######################################################################
#                    Initialdata entropy variables                    #
#######################################################################


function initialdata_entropy_variables!(::Val, P, prms)
  TODO(initialdata_entropy_variables!)
end

#######################################################################
#                          InitialData HRSC                           #
#######################################################################

initialdata_hrsc!(kind, env, P, hrsc::Nothing, prms) = nothing
function initialdata_hrsc!(kind, env, P, hrsc, prms)
  TODO(initialdata_hrsc!)
end

function initialdata_hrsc!(kind, env, P::Project, hrsc::HRSC.AbstractArtificialViscosity, prms)
  @unpack mesh = env
  @unpack mu = get_cell_variables(mesh)
  @. mu = 0.0
end

function initialdata_hrsc!(kind, env, P::Project, hrsc::HRSC.SmoothedArtificialViscosity, prms)
  @unpack mesh = env
  @unpack mu = get_cell_variables(mesh)
  @unpack smoothed_mu = get_static_variables(mesh)
  initialdata_hrsc!(kind, env, P, hrsc.av, prms)
  hrsc.smoother(smoothed_mu, mu, env.mesh, mesh.tree.periodic[1])
end

#######################################################################
#                           InitialData TCI                           #
#######################################################################

function initialdata_tci!(::Val, P, tci, prms)
  TODO(initialdata_tci!)
end
