#######################################################################
#                              Timestep                               #
#######################################################################

function compute_maxspeed(::Project{:spherical1d}, equation, mesh)
  broadcast_volume!(maxspeed, equation, mesh)
end
function compute_maxspeed(::Project{:rescaled_spherical1d}, equation, mesh)
  broadcast_volume!(maxspeed_rescaled_spherical1d, equation, mesh)
end

function get_maxspeed(mesh, equation)
  @unpack max_v = get_static_variables(mesh.cache)
  vmax = maximum(abs, max_v)
  vmax_limit = 0.99999999999
  if vmax > vmax_limit
    @warn "Limiting timestep due to maximum speed exceeding $vmax_limit"
    vmax = vmax_limit
  end
  return vmax
end

function timestep(mesh, P::Project, hrsc::Maybe{HRSC.AbstractHRSC})
  compute_maxspeed(P, P.equation, mesh)
  vmax = get_maxspeed(mesh, P.equation)
  dl = min_grid_spacing(mesh)
  dt = dl / (vmax * dtfactor(mesh))
  return dt
end
dtfactor(mesh::Mesh1d{FVElement}) = 2
dtfactor(mesh::Mesh1d{SpectralElement}) = mesh.element.N


function timestep(mesh, P::Project, hrsc::HRSC.AbstractArtificialViscosity)
  @unpack equation = P

  compute_maxspeed(P, P.equation, mesh)
  vmax = get_maxspeed(mesh, P.equation)

  @unpack mu = get_cell_variables(mesh)
  mumax = maximum(abs, mu)

  @unpack element = mesh
  @unpack N = element
  L, = dg1d.widths(mesh)
  K, = mesh.tree.dims
  dL = L/K
  dl = min_grid_spacing(mesh)

  dt1 = dl / (vmax * N)
  dt2 = dL^2 / (mumax * N^3)
  dt = min(dt1, dt2)
  dt_limit = 1e-7
  if dt < dt_limit
    @warn "Limiting timestep due to dt being smaller than $dt_limit"
    dt = dt_limit
  end
  return dt
end


#######################################################################
#                           RHS computation                           #
#######################################################################


function rhs!(mesh::Mesh1d{FVElement}, P::Project{:spherical1d}, hrsc::Nothing)

  @unpack cache = mesh
  @unpack equation = P

  @unpack D, Sr, τ                = get_dynamic_variables(cache)
  @unpack flr_D, flr_Sr, flr_τ,
          max_v, ρ, ϵ, vr, p,
          src_D, src_Sr, src_τ    = get_static_variables(cache)
  @unpack rhs_D, rhs_Sr, rhs_τ    = get_rhs_variables(cache)
  @unpack nflr_D, nflr_Sr, nflr_τ,
          bdry_D, bdry_Sr, bdry_τ = get_bdry_variables(cache)
  @unpack dt                      = get_global_variables(cache)

  broadcast_volume!(cons2prim, equation, mesh)
  broadcast_volume!(maxspeed, equation, mesh)
  broadcast_volume!(flux_source_spherical1d, equation, mesh)
  broadcast_bdry!(fv_bdry_flux, equation, mesh)

  fv_update_step!(rhs_D,  D,  flr_D,  src_D,  bdry_D,  nflr_D,  dt[1], mesh)
  fv_update_step!(rhs_Sr, Sr, flr_Sr, src_Sr, bdry_Sr, nflr_Sr, dt[1], mesh)
  fv_update_step!(rhs_τ,  τ,  flr_τ,  src_τ,  bdry_τ,  nflr_τ,  dt[1], mesh)

  return
end


function rhs!(mesh::Mesh1d, P::Project{:spherical1d}, hrsc::Nothing)

  @unpack cache = mesh
  @unpack equation = P

  @unpack D, Sr, τ                    = get_dynamic_variables(cache)
  @unpack rhs_D, rhs_Sr, rhs_τ        = get_rhs_variables(cache)
  @unpack flr_D, flr_Sr, flr_τ,
          max_v, vr, p,
          src_D, src_Sr, src_τ        = get_static_variables(cache)
  @unpack nflr_D, nflr_Sr, nflr_τ,
          bdry_D, bdry_Sr, bdry_τ,
          bdry_max_v, bdry_vr, bdry_p = get_bdry_variables(cache)

  broadcast_volume!(cons2prim, equation, mesh)
  broadcast_volume!(maxspeed, equation, mesh)

  dg1d.interpolate_face_data!(mesh, D,     bdry_D)
  dg1d.interpolate_face_data!(mesh, Sr,    bdry_Sr)
  dg1d.interpolate_face_data!(mesh, τ,     bdry_τ)
  dg1d.interpolate_face_data!(mesh, max_v, bdry_max_v)
  dg1d.interpolate_face_data!(mesh, vr,    bdry_vr)
  dg1d.interpolate_face_data!(mesh, p,     bdry_p)

  broadcast_volume!(flux_source_spherical1d, equation, mesh)
  impose_symmetry_sources!(P, mesh)
  broadcast_faces!(llf, equation, mesh)
  broadcast_bdry!(bdryllf, equation, mesh)

  compute_rhs_weak_form!(rhs_D,  flr_D,  src_D,  nflr_D,  mesh)
  compute_rhs_weak_form!(rhs_Sr, flr_Sr, src_Sr, nflr_Sr, mesh)
  compute_rhs_weak_form!(rhs_τ,  flr_τ,  src_τ,  nflr_τ,  mesh)

  return
end


function rhs!(mesh::Mesh1d{FVElement}, P::Project{:rescaled_spherical1d}, hrsc::Nothing)

  @unpack cache = mesh
  @unpack equation = P

  @unpack rD, rSr, rτ                = get_dynamic_variables(cache)
  @unpack flr_rD, flr_rSr, flr_rτ,
          D, Sr, τ,
          src_rD, src_rSr, src_rτ    = get_static_variables(cache)
  @unpack rhs_rD, rhs_rSr, rhs_rτ    = get_rhs_variables(cache)
  @unpack nflr_rD, nflr_rSr, nflr_rτ,
          bdry_D, bdry_Sr, bdry_τ = get_bdry_variables(cache)

  broadcast_volume!(unscale_conservatives, equation, mesh)
  impose_symmetry!(P, mesh)
  broadcast_volume!(cons2prim_rescaled_spherical1d, equation, mesh)
  broadcast_volume!(maxspeed_rescaled_spherical1d, equation, mesh)
  broadcast_volume!(flux_source_rescaled_spherical1d, equation, mesh)
  broadcast_bdry!(fv_bdry_flux_rescaled_spherical1d, equation, mesh)

  dt = timestep(mesh, P, hrsc)
  fv_update_step!(rhs_rD,  rD,  flr_rD,  src_rD,  bdry_D,  nflr_rD,  dt, mesh)
  fv_update_step!(rhs_rSr, rSr, flr_rSr, src_rSr, bdry_Sr, nflr_rSr, dt, mesh)
  fv_update_step!(rhs_rτ,  rτ,  flr_rτ,  src_rτ,  bdry_τ,  nflr_rτ,  dt, mesh)

  return
end


function rhs!(mesh::Mesh1d, P::Project{:rescaled_spherical1d}, hrsc::Nothing)

  @unpack cache = mesh
  @unpack equation = P

  @unpack rD, rSr, rτ                    = get_dynamic_variables(cache)
  @unpack rhs_rD, rhs_rSr, rhs_rτ        = get_rhs_variables(cache)
  @unpack flr_rD, flr_rSr, flr_rτ,
          max_v, vr, p, D, Sr, τ,
          src_rD, src_rSr, src_rτ        = get_static_variables(cache)
  @unpack nflr_rD, nflr_rSr, nflr_rτ,
          bdry_D, bdry_Sr, bdry_τ,
          bdry_max_v, bdry_vr, bdry_p = get_bdry_variables(cache)

  broadcast_volume!(unscale_conservatives, equation, mesh)
  impose_symmetry!(P, mesh)
  broadcast_volume!(cons2prim_rescaled_spherical1d, equation, mesh)
  broadcast_volume!(maxspeed_rescaled_spherical1d, equation, mesh)

  dg1d.interpolate_face_data!(mesh, D,     bdry_D)
  dg1d.interpolate_face_data!(mesh, Sr,    bdry_Sr)
  dg1d.interpolate_face_data!(mesh, τ,     bdry_τ)
  dg1d.interpolate_face_data!(mesh, max_v, bdry_max_v)
  dg1d.interpolate_face_data!(mesh, vr,    bdry_vr)
  dg1d.interpolate_face_data!(mesh, p,     bdry_p)

  broadcast_volume!(flux_source_rescaled_spherical1d, equation, mesh)
  broadcast_faces!(llf_rescaled_spherical1d, equation, mesh)
  broadcast_bdry!(bdryllf_rescaled_spherical1d, equation, mesh)

  compute_rhs_weak_form!(rhs_rD,  flr_rD,  src_rD,  nflr_rD,  mesh)
  compute_rhs_weak_form!(rhs_rSr, flr_rSr, src_rSr, nflr_rSr, mesh)
  compute_rhs_weak_form!(rhs_rτ,  flr_rτ,  src_rτ,  nflr_rτ,  mesh)

  return
end


function rhs!(env, P::Project, hrsc::HRSC.AbstractReconstruction)

  TODO()

  @unpack cache, mesh = env
  @unpack equation, rsolver = P

  @unpack sD, sSr, stau                    = get_dynamic_variables(cache)
  @unpack flx_sD, flx_sSr, flx_stau        = get_static_variables(cache)
  @unpack rhs_sD, rhs_sSr, rhs_stau        = get_rhs_variables(cache)
  @unpack lhs_numflx_sD,   rhs_numflx_sD,
          lhs_numflx_sSr,  rhs_numflx_sSr,
          lhs_numflx_stau, rhs_numflx_stau = get_bdry_variables(cache)
  @unpack flag                             = get_cell_variables(cache)

  HRSC.reconstruct!(sD,    flag, hrsc, isperiodic=isperiodic(bdryconds), m=1e-10, limit_with_boundaries=false)
  HRSC.reconstruct!(sSr,   flag, hrsc, isperiodic=isperiodic(bdryconds), limit_with_boundaries=false)
  HRSC.reconstruct!(stau,  flag, hrsc, isperiodic=isperiodic(bdryconds), m=1e-10, limit_with_boundaries=false)

  broadcast_volume!(scaledcons2cons, equation, cache)
  postprocess_scaledcons2cons!(cache, equation, mesh)
  broadcast_volume!(cons2prim, equation, cache)
  broadcast_volume!(flux, equation, cache)
  broadcast_volume!(sources, equation, cache)
  broadcast_faces!(lax_friedrich_flux, rsolver, cache, mesh)

  compute_rhs_weak_form!(rhs_sD,   flx_sD,    lhs_numflx_sD,   rhs_numflx_sD,   mesh)
  compute_rhs_weak_form!(rhs_sSr,  flx_sSr,   lhs_numflx_sSr,  rhs_numflx_sSr,  mesh)
  compute_rhs_weak_form!(rhs_stau, flx_stau,  lhs_numflx_stau, rhs_numflx_stau, mesh)

  return
end


function rhs!(mesh, P::Project{:spherical1d}, hrsc::HRSC.AbstractArtificialViscosity)

  @unpack cache = mesh
  @unpack equation = P

  @unpack D, Sr, τ                            = get_dynamic_variables(cache)
  @unpack rhs_D, rhs_Sr, rhs_τ                = get_rhs_variables(cache)
  @unpack flr_D, flr_Sr, flr_τ,
          ldg_D, ldg_Sr, ldg_τ,
          max_v, vr, p, smoothed_mu,
          src_D, src_Sr, src_τ                = get_static_variables(cache)
  @unpack nflr_D, nflr_Sr, nflr_τ,
          bdry_D, bdry_Sr, bdry_τ,
          bdry_max_v, bdry_vr, bdry_p,
          bdry_ldg_D, bdry_ldg_Sr, bdry_ldg_τ,
          bdry_rhs_D, bdry_rhs_Sr, bdry_rhs_τ,
          bdry_smoothed_mu                    = get_bdry_variables(cache)

  broadcast_volume!(cons2prim, equation, mesh)
  broadcast_volume!(maxspeed, equation, mesh)

  dg1d.interpolate_face_data!(mesh, D,     bdry_D)
  dg1d.interpolate_face_data!(mesh, Sr,    bdry_Sr)
  dg1d.interpolate_face_data!(mesh, τ,     bdry_τ)
  dg1d.interpolate_face_data!(mesh, max_v, bdry_max_v)
  dg1d.interpolate_face_data!(mesh, vr,    bdry_vr)
  dg1d.interpolate_face_data!(mesh, p,     bdry_p)

  broadcast_volume!(flux_source_spherical1d, equation, mesh)
  impose_symmetry_sources!(P, mesh)
  broadcast_faces!(llf, equation, mesh)
  broadcast_bdry!(bdryllf, equation, mesh)

  compute_rhs_weak_form!(rhs_D,  flr_D,  src_D,  nflr_D,  mesh)
  compute_rhs_weak_form!(rhs_Sr, flr_Sr, src_Sr, nflr_Sr, mesh)
  compute_rhs_weak_form!(rhs_τ,  flr_τ,  src_τ,  nflr_τ,  mesh)

  # now compute rhs for ∂t u + ∂x f(x) = ∂x( μ (-∂t+∂x) u )

  broadcast_faces!(ldg_nflux_spherical1d, equation, mesh)
  broadcast_bdry!(ldg_bdryllf_spherical1d, equation, mesh)
  compute_rhs_weak_form!(ldg_D,  D,  nflr_D,  mesh)
  compute_rhs_weak_form!(ldg_Sr, Sr, nflr_Sr, mesh)
  compute_rhs_weak_form!(ldg_τ,  τ,  nflr_τ,  mesh)

  # need to recompute, because ldg computation overwrites nflx_{D,Sr,tau}
  # numerical fluxes
  broadcast_faces!(llf, equation, mesh)
  broadcast_bdry!(bdryllf, equation, mesh)

  broadcast_volume!(av_flux_covariant_spherical1d, equation, mesh)

  dg1d.interpolate_face_data!(mesh, ldg_D,       bdry_ldg_D)
  dg1d.interpolate_face_data!(mesh, ldg_Sr,      bdry_ldg_Sr)
  dg1d.interpolate_face_data!(mesh, ldg_τ,       bdry_ldg_τ)
  dg1d.interpolate_face_data!(mesh, rhs_D,       bdry_rhs_D)
  dg1d.interpolate_face_data!(mesh, rhs_Sr,      bdry_rhs_Sr)
  dg1d.interpolate_face_data!(mesh, rhs_τ,       bdry_rhs_τ)
  dg1d.interpolate_face_data!(mesh, smoothed_mu, bdry_smoothed_mu)

  broadcast_faces!(av_nflux_covariant_spherical1d, equation, mesh)

  compute_rhs_weak_form!(rhs_D,  flr_D,  src_D,  nflr_D,  mesh)
  compute_rhs_weak_form!(rhs_Sr, flr_Sr, src_Sr, nflr_Sr, mesh)
  compute_rhs_weak_form!(rhs_τ,  flr_τ,  src_τ,  nflr_τ,  mesh)

  return
end


# function rhs!(mesh, P::Project, hrsc::HRSC.AbstractArtificialViscosity)
#
#   @unpack D, Sr, tau                          = get_dynamic_variables(mesh)
#   @unpack flx_D, flx_Sr, flx_tau,
#           ldg_D, ldg_Sr, ldg_tau = get_static_variables(mesh)
#   @unpack rhs_sD, rhs_sSr, rhs_stau              = get_rhs_variables(mesh)
#
#   broadcast_volume!(scaledcons2cons, equation, cache)
#   postprocess_scaledcons2cons!(cache, equation, mesh)
#   # broadcast_volume!(conservative_fixing, equation, cache)
#   broadcast_volume!(cons2prim, equation, cache)
#   # broadcast_volume!(atmosphering, equation, cache)
#   broadcast_volume!(cons2scaledcons, equation, cache)
#
#   ## solve auxiliary equation: q + ∂x u = 0
#   broadcast_volume!(ldg_flux, equation, cache)
#   broadcast_faces!(mean_flux, ldg_rsolver, cache, mesh)
#   broadcast_boundaryconditions!(mean_flux, ldg_bdryconds, cache, mesh, 0.0)
#   compute_rhs_weak_form!(ldg_sD,   flx_ldg_sD,   lhs_numflx_ldg_sD,   rhs_numflx_ldg_sD,   mesh)
#   compute_rhs_weak_form!(ldg_sSr,  flx_ldg_sSr,  lhs_numflx_ldg_sSr,  rhs_numflx_ldg_sSr,  mesh)
#   compute_rhs_weak_form!(ldg_stau, flx_ldg_stau, lhs_numflx_ldg_stau, rhs_numflx_ldg_stau, mesh)
#
#   ## compute rhs of regularized equation: ∂t u + ∂x f + ∂x μ q = 0
#   broadcast_volume!(av_flux, equation, cache)
#   broadcast_faces!(mean_flux, av_rsolver, cache, mesh)
#   broadcast_boundaryconditions!(mean_flux, av_bdryconds, cache, mesh, 0.0)
#
#   # broadcast_volume!(scaledcons2cons, equation, cache)
#   # postprocess_scaledcons2cons!(cache, equation, mesh)
#   broadcast_volume!(flux, equation, cache)
#   broadcast_volume!(sources, equation, cache)
#   broadcast_faces!(lax_friedrich_flux, rsolver, cache, mesh)
#   broadcast_boundaryconditions!(lax_friedrich_flux, bdryconds, cache, mesh, 0.0)
#
#   @. flx_sD          += flx_ldg_sD
#   @. flx_sSr         += flx_ldg_sSr
#   @. flx_stau        += flx_ldg_stau
#   @. lhs_numflx_sD   += lhs_numflx_ldg_sD
#   @. lhs_numflx_sSr  += lhs_numflx_ldg_sSr
#   @. lhs_numflx_stau += lhs_numflx_ldg_stau
#   @. rhs_numflx_sD   += rhs_numflx_ldg_sD
#   @. rhs_numflx_sSr  += rhs_numflx_ldg_sSr
#   @. rhs_numflx_stau += rhs_numflx_ldg_stau
#
#   compute_rhs_weak_form!(rhs_sD,   flx_sD,   lhs_numflx_sD,   rhs_numflx_sD,   mesh)
#   compute_rhs_weak_form!(rhs_sSr,  flx_sSr,  lhs_numflx_sSr,  rhs_numflx_sSr,  mesh)
#   compute_rhs_weak_form!(rhs_stau, flx_stau, lhs_numflx_stau, rhs_numflx_stau, mesh)
#
#   return true
# end
