function Project(env::Environment, prms)

  # construct Project
  eos = EquationOfState.make_EquationOfState(prms["EquationOfState"])
  equation = make_Equation(eos, prms["GRHD"], env.mesh)
  tci  = TCI.make_TCI(env.mesh, prms["TCI"])
  hrsc = HRSC.make_HRSC(env.mesh, prms["HRSC"])

  fixedprms = (;)
  formulation = prms["GRHD"]["formulation"]
  P = Project(Symbol(formulation), equation, hrsc, tci, fixedprms)

  # register variables
  @unpack cache = env
  register!(cache, P)
  display(cache)

  # setup initial data
  initialdata!(env, P, prms["GRHD"])

  # setup callbacks
  projectcb = make_callback(env, P, isperiodic(env.mesh))

  append!(env.callbacks, CallbackSet(projectcb.callbacks...))

  return P, nothing
end


#######################################################################
#                         Register variables                          #
#######################################################################


function register!(cache, P::Project{:spherical1d})
  register_evolution_spherical1d!(cache)
  register_metricvars_spherical1d!(cache)
  register_hrsc_spherical1d!(cache, P.hrsc)
  register_tci_spherical1d!(cache, P.tci)
end
function register!(cache, P::Project{:rescaled_spherical1d})
  register_evolution_rescaled_spherical1d!(cache)
  register_metricvars_rescaled_spherical1d!(cache)
  register_hrsc_rescaled_spherical1d!(cache, P.hrsc)
  register_tci_rescaled_spherical1d!(cache, P.tci)
end


function register_evolution_spherical1d!(cache)
  register_variables!(cache,
    dynamic_variablenames = (:D, :Sr, :τ,),
    rhs_variablenames     = (:rhs_D, :rhs_Sr, :rhs_τ), # rhs of conservatives
    static_variablenames  = (:flr_D, :flr_Sr, :flr_τ,  # conservatives' fluxes
                             :src_D, :src_Sr, :src_τ,  # sources
                             :ρ, :vr, :p, :ϵ,          # primitives
                             :max_v, :r,               # maximum charactersitic speed, radius
                             :init_D, :init_Sr, :init_τ, :init_max_v, :init_vr, :init_p,
                             :E, :Em1, :flr_E, :flr_Em1, :EP),
    bdry_variablenames    = (:bdry_D, :bdry_Sr, :bdry_τ,
                             :bdry_ρ, :bdry_ϵ, :bdry_vr, :bdry_p,
                             :bdry_max_v,
                             :nflr_D, :nflr_Sr, :nflr_τ),
    cell_variablenames    = (:cellmax_v,),
    global_variablenames  = (:t, :tm1) # time steps
  )
  @unpack r, x = get_static_variables(cache)
  r .= x
end
function register_evolution_rescaled_spherical1d!(cache)
  register_variables!(cache,
    dynamic_variablenames = (:rD, :rSr, :rτ,),            # we evolve rescaled variables
    rhs_variablenames     = (:rhs_rD, :rhs_rSr, :rhs_rτ), # rhs of conservatives
    static_variablenames  = (:flr_rD, :flr_rSr, :flr_rτ,  # conservatives' fluxes
                             :src_rD, :src_rSr, :src_rτ,  # sources
                             :D, :Sr, :τ,                 # conserved variables
                             :ρ, :vr, :p, :ϵ,             # primitives
                             :max_v, :r,                  # maximum charactersitic speed, radius
                             :init_D, :init_Sr, :init_τ, :init_max_v, :init_vr, :init_p,
                             :E, :Em1, :flr_E, :flr_Em1, :EP),
    bdry_variablenames    = (:bdry_D, :bdry_Sr, :bdry_τ,
                             :bdry_ρ, :bdry_ϵ, :bdry_vr, :bdry_p,
                             :bdry_max_v,
                             :nflr_rD, :nflr_rSr, :nflr_rτ),
    cell_variablenames    = (:cellmax_v,),
    global_variablenames  = (:t, :tm1) # time steps
  )
  @unpack r, x = get_static_variables(cache)
  r .= x
end


function register_metricvars_spherical1d!(cache)
  register_variables!(cache,
    static_variablenames = (:α, :βru,
                            :γrr, :γθθ, :γϕϕ,
                            :Krr, :Kθθ, :Kϕϕ,
                            :Γrrr, :Γrθθ, :Γrϕϕ, :Γθrθ, :Γθϕϕ, :Γϕrϕ, :Γϕθϕ,
                            :∂αr, :∂βrrdu,
                            :∂γrrr, :∂γrθθ, :∂γrϕϕ, :∂γθϕϕ,)
  )
end
function register_metricvars_rescaled_spherical1d!(cache)
  register_variables!(cache,
    static_variablenames = (:A, :∂Ar, :B, :∂Br)
  )
end


# TODO Where to put these? Can be needed by either HRSC or TCI.
# :E, :Em1,                   # entropy
# :flx_E, :flx_Em1,           # entropy flux


#######################################################################
#                                HRSC                                 #
#######################################################################

register_hrsc_spherical1d!(cache, hrsc::Nothing) = nothing
register_hrsc_rescaled_spherical1d!(cache, hrsc::Nothing) = nothing
register_hrsc_spherical1d!(cache, hrsc::AbstractHRSC) = TODO(hrsc)
register_hrsc_rescaled_spherical1d!(cache, hrsc::AbstractHRSC) = TODO(hrsc)

function register_hrsc_spherical1d!(cache, hrsc::HRSC.AbstractArtificialViscosity)
  register_variables!(cache,
    static_variablenames  = (:ldg_D, :ldg_Sr, :ldg_τ),
    bdry_variablenames    = (:bdry_ldg_D, :bdry_ldg_Sr, :bdry_ldg_τ,
                             :bdry_smoothed_mu,
                             :bdry_rhs_D, :bdry_rhs_Sr, :bdry_rhs_τ),
    cell_variablenames    = (:mu, :tmp_mu), # 'one viscosity to rule them all'
    )
end
function register_hrsc_spherical1d!(cache, hrsc::HRSC.SmoothedArtificialViscosity)
  register_hrsc_spherical1d!(cache, hrsc.av)
  register_variables!(cache,
    static_variablenames  = (:smoothed_mu,)
  )
end


#######################################################################
#                                 TCI                                 #
#######################################################################


register_tci_spherical1d!(cache, tci::Maybe{HRSC.AbstractReconstruction}) = nothing
register_tci_rescaled_spherical1d!(cache, tci::Maybe{HRSC.AbstractReconstruction}) = nothing

function register_tci_spherical1d!(cache, tci::TCI.AbstractTCI)
  TODO(tci)
  register_variables!(cache, cell_variablenames = (:flag,:sD_flag,:sSr_flag,:stau_flag))
end
function register_tci_spherical1d!(cache, tci::TCI.EntropyProduction)
  TODO(tci)
  register_variables!(cache,
                      cell_variablenames = (:flag,),
                      static_variablenames = (:EP,), # entropy production
                      bdry_variablenames   = (:EPjump_lhs, :EPjump_rhs)
                                              # entropy production due to jumps at interfaces
  )
end
