module IntegrationTests

using dg1d

using Printf
using PrettyTables
using LinearAlgebra


const REFDIR = abspath(joinpath(@__DIR__, "..", "refs"))
const TESTDIR = abspath(joinpath(@__DIR__, "..", "tests"))
const TERM_GREEN = "\033[32m"
const TERM_BLINK_RED = isinteractive() ? "\033[5m\033[31m" : ""
const TERM_RESET = isinteractive() ? "\033[0m" : ""


function my_findmax(f::Function, args...)
    max, idx = f(first.(args)...), 1
    for ii in eachindex(args...)
        ai = ( a[ii] for a in args )
        fi = f(ai...)
        if fi > max
            max = fi
            idx = ii
        end
    end
    return max, idx
end


"""
    discover_output_dir(outdir)

A directory named `outdir` is an output directory if
- it contains a parameter file name `outdir.toml`
- it contains an output file name `output.h5`
"""
function load_reftest(testname)

    outdir = joinpath(REFDIR, testname)
    if !isdir(outdir)
        error("No output directory found for reference test '$testname'")
    end

    _, _, files = first(walkdir(outdir))

    parfile = "$testname.toml"
    if !(parfile in files)
        error("No parameter file '$parfile' found in output directory for reference test '$testname'")
    end

    outfile = "output.h5"
    if !(outfile in files)
        error("No output file '$outfile' found in output directory for reference test '$testname'")
    end

    intrpfile = "interpolate.h5"
    # this file is optional, so we don't error

    return outdir, parfile, outfile, intrpfile
end


function compare_outputs(io, testname, refoutput, output, variables, reltol, abstol)::Bool

    rout = try
        dg1d.Output(refoutput)
    catch e
        println(io, """
                    FAIL: Failed to open reference output '$refoutput'
                    """)
        return false
    end
    out = try
        dg1d.Output(output)
    catch e
        println(io, """
                    FAIL: Failed to open output '$output'
                    """)
        return false
    end

    ref_x, ref_y = rout.x, rout.y
    out_x, out_y = out.x, out.y
    if length(ref_x) != length(out_x)
        println(io, """
                    FAIL: Mismatch in mesh size
                    reference $(length(ref_x))
                    test      $(length(out_x))
                    """)
        return false
    end
    if typeof(ref_y) != typeof(out_y) || length(ref_y) != length(out_y)
        println(io, """
                    FAIL: Mismatch in mesh size
                    reference $(length(ref_y))
                    test      $(length(out_y))
                    """)
        return false
    end

    refts = rout.times
    ts = out.times
    if length(refts) != length(ts)
        println(io, """
                    FAIL: Mismatch in number of time steps for which outputs were recorded
                    """)
        nref, ntest= length(refts), length(ts)
        common_n = max(nref,ntest)
        data = zeros(Union{Float64,Missing}, common_n, 2)
        data .= missing
        data[1:nref,1]  .= refts
        data[1:ntest,2] .= ts
        pretty_table(io, data;
            header = [ :ref, :test ],
            tf = tf_simple,
            formatters = ft_printf("% .6e"),
            crop = :none)
        return false
    end

    if any(zip(refts, ts)) do (rti, ti)
            !isapprox(rti, ti)
        end

        println(io, "FAIL: Mismatch in output times:\n")
        data = hcat(refts, ts)
        pretty_table(io, data;
            header = [ :reference, :test ],
            tf = tf_simple,
            formatters = ft_printf("% 10.10e"),
            crop = :none)
        return false
    end

    success = true
    for var in variables
        if !dg1d.has_variable(rout, var)
            println(io, "FAIL: Missing variable '$var' in reference output")
            success = false
            continue
        end
        if !dg1d.has_variable(out, var)
            println(io, "FAIL: Missing variable '$var' in test output")
            success = false
            continue
        end

        check_failed = false
        diff_ts, diff_xs, diff_is, diff_maxs = Float64[], Float64[], Float64[], Float64[]
        for i in 1:length(ts)
            refvi = read_variable(rout, var, i)
            vi = read_variable(out, var, i)

            if !all(zip(refvi, vi)) do (rv, v)
                    isapprox(rv, v, rtol=reltol, atol=abstol)
                end
                max, idx = my_findmax(refvi,vi) do rv,v
                    abs(rv-v)
                end
                push!(diff_is, i)
                push!(diff_ts, ts[i])
                push!(diff_xs, ref_x[idx])
                push!(diff_maxs, max)
                success = false
                check_failed = true
            end
        end
        if check_failed
            println(io, "FAIL: Mismatch in variable '$var'")
            pretty_table(io, hcat(diff_is, diff_ts, diff_xs, diff_maxs);
                         header = [ "i", "t[i]", "x", "max(abs(ref-test))" ],
                         tf = tf_simple,
                         formatters = ft_printf(["%d", "% .6e", "% .6e", "% .6e"],1:4),
                         crop = :none)
            println(io)

            # copy outputs to test/diffoutputs/testname which is then uploaded to CI
            diffdir = joinpath(TESTDIR, "diffoutputs")
            if !ispath(diffdir)
                mkdir(diffdir)
            end
            testdir = joinpath(diffdir, testname)
            if !ispath(testdir)
                mkdir(testdir)
            end
            _, ext = splitext(refoutput)
            cp(refoutput, joinpath(testdir, "refoutput$ext"), force=true)
            _, ext = splitext(output)
            cp(output, joinpath(testdir, "output$ext"), force=true)
        end
    end

    return success
end


function load_testconfig()

    filename = normpath(joinpath(REFDIR, "testconfig.toml"))
    testconfig = TOML.parsefile(filename)

    # checks for global settings
    default_reltol = 1e-8
    default_abstol = 1e-10
    default_intrp_reltol = 1e-8
    default_intrp_abstol = 1e-10
    default_reltol = get!(testconfig, "default_reltol", default_reltol)
    default_abstol = get!(testconfig, "default_abstol", default_abstol)
    if default_reltol < 0
        error("Invalid testconfig.toml: 'default_reltol' must be positive, found 'default_reltol = $default_reltol'")
    end
    if default_abstol < 0
        error("Invalid testconfig.toml: 'default_abstol' must be positive, found 'default_abstol = $default_abstol'")
    end
    default_intrp_reltol = get!(testconfig, "default_interpolate_reltol", default_intrp_reltol)
    default_intrp_abstol = get!(testconfig, "default_interpolate_abstol", default_intrp_abstol)
    if default_intrp_reltol < 0
        error("Invalid testconfig.toml: 'default_interpolate_reltol' must be positive, found 'default_interpolate_reltol = $default_intrp_reltol'")
    end
    if default_intrp_abstol < 0
        error("Invalid testconfig.toml: 'default_interpolate_abstol' must be positive, found 'default_interpolate_abstol = $default_intrp_abstol'")
    end

    # drop default tolerance parameters
    testconfig = filter(testconfig) do (k,_)
        !(k in ["default_reltol", "default_abstol", "default_interpolate_reltol", "default_interpolate_abstol"])
    end
    # display(testconfig)

    reftests = String[]
    for (registered_test, pars) in testconfig
        if !(pars isa AbstractDict)
            error("Invalid testconfig.toml: Found unsupported parameter '$registered_test'")
        end
        # if no vars are given, we compare all available ones
        vars   = get!(pars, "variables", String[])
        reltol = get!(pars, "reltol", default_reltol)
        abstol = get!(pars, "abstol", default_abstol)
        if default_reltol < 0
            error("Invalid testconfig.toml: '$registered_test.reltol' must be positive, found '$registered_test.reltol = $reltol'")
        end
        if default_abstol < 0
            error("Invalid testconfig.toml: '$registered_test.abstol' must be positive, found '$registered_test.abstol = $abstol'")
        end
        # if no vars are given, we compare all available ones
        intrpvars   = get!(pars, "interpolate_variables", String[])
        intrp_reltol = get!(pars, "interpolate_reltol", default_intrp_reltol)
        intrp_abstol = get!(pars, "interpolate_abstol", default_intrp_abstol)
        if default_intrp_reltol < 0
            error("Invalid testconfig.toml: '$registered_test.interpolate_reltol' must be positive, found '$registered_test.interpolate_reltol = $intrp_reltol'")
        end
        if default_intrp_abstol < 0
            error("Invalid testconfig.toml: '$registered_test.interpolate_abstol' must be positive, found '$registered_test.interpolate_abstol = $intrp_abstol'")
        end

        push!(reftests, registered_test)
    end

    return testconfig

end


"""
    runtests(tests::Vector{String})

Run integration tests for which there is reference data
in `refs/` and a configuration entry in `refs/testconfig.toml`.
"""
function runtests(tests::Vector{String}=String[])

    ### load testconfig and gather files

    testconfig = load_testconfig()

    # gather all reference tests
    allreftests = collect(keys(testconfig))
    allrefdirs, allrefpars, allrefoutputs, allrefintrpfiles = String[], String[], String[], Union{Nothing,String}[]
    for reftest in allreftests
        outdir, parfile, outfile, intrpfile = load_reftest(reftest)
        push!(allrefdirs, outdir)
        push!(allrefpars, parfile)
        push!(allrefoutputs, outfile)
        push!(allrefintrpfiles, intrpfile)
    end

    # select dirs, parfiles and outputs for requested tests
    if length(tests) == 0
        # run all available tests
        tests = allreftests
    else
        foreach(tests) do t
            if !(t in allreftests)
                error("Unknown test '$t' request. Make sure there is a folder named '$t' in 'refs/' and a configuration entry in 'refs/testconfig.toml'")
            end
        end
    end
    idxs = [ findfirst(isequal(t), allreftests) for t in tests ]
    refdirs = allrefdirs[idxs]
    refpars = allrefpars[idxs]
    refoutputs = allrefoutputs[idxs]
    refintrpfiles = allrefintrpfiles[idxs]

    ### prepare tests

    # clean up any previous tests
    if isdir(TESTDIR)
        rm(TESTDIR, recursive=true, force=true)
    end
    mkdir(TESTDIR)

    # copy reference parfiles to tests directory
    newpars = String[]
    for (refdir, refpar) in zip(refdirs, refpars)
        rpar = joinpath(REFDIR, refdir, refpar)
        par = joinpath(TESTDIR, refpar)
        cp(rpar, par)
        push!(newpars, par)
    end
    newdirs = [ first(splitext(p)) for p in newpars ]
    newoutputs = [ out for out in refoutputs ]
    newintrpfiles = [ out for out in refintrpfiles ]

    ### run the tests

    runs_status = Union{Tuple{Any#=Exception=#,String},Nothing}[]
    for p in newpars
        @info "Running parameter file '$p' ..."
        ex_msg = try
            @noinline dg1d.main(p)
            nothing
        catch e
            (e, sprint(showerror, e, catch_backtrace()))
        end
        push!(runs_status, ex_msg)
    end

    ### run comparsions between reference result and new outputs

    logs = IOBuffer()
    errlogs = IOBuffer()
    tmplogs = IOBuffer()
    n_header = 80
    comparsion_success = Union{Bool,Missing}[]
    for i in 1:length(tests)

        test = tests[i]
        println(logs, repeat('+', n_header), "\n\nTest: $test\n")

        status = runs_status[i]
        success = isnothing(status)
        if !success
            e, msg = status
            println(logs, "FAIL: Failed to run test!")
            println(logs, "      $(summary(e))")
            push!(comparsion_success, missing)
            println(logs, "\nResult: ", TERM_BLINK_RED, "FAILED", TERM_RESET, "\n")
            println(errlogs, "="^100)
            println(errlogs, "Reported by test $test")
            println(errlogs, "      $msg")
            continue
        end

        refout = joinpath(REFDIR, refdirs[i], refoutputs[i])
        newout = joinpath(TESTDIR, newdirs[i], newoutputs[i])
        intrp_refout = joinpath(REFDIR, refdirs[i], refintrpfiles[i])
        intrp_newout = joinpath(TESTDIR, newdirs[i], newintrpfiles[i])
        @unpack variables, reltol, abstol,
                interpolate_variables, interpolate_reltol, interpolate_abstol = testconfig[test]
        comp_success = compare_outputs(tmplogs, test, refout, newout, variables, reltol, abstol)
        if !comp_success
            # merge logs
            println(logs, "Details comparison output.h5\n")
            println(logs, String(take!(tmplogs)))
        end
        if isfile(intrp_refout)
            # interpolate.h5 is optional test output
            intrp_comp_success = compare_outputs(tmplogs, test, intrp_refout, intrp_newout,
                                                 interpolate_variables, interpolate_reltol, interpolate_abstol)
            if !intrp_comp_success
                # merge logs
                println(logs, "Details comparsion interpolate.h5\n")
                println(logs, String(take!(tmplogs)))
            end
            comp_success = comp_success && intrp_comp_success
        end
        push!(comparsion_success, comp_success)

        result = success && comp_success

        println(logs, result ? "" : "\n", "Result: ",
                result ? TERM_GREEN*"PASSED" : TERM_BLINK_RED*"FAILED", TERM_RESET, "\n")
    end

    ### print summary

    n_runs             = length(runs_status)
    n_runs_succeeded   = count(isnothing, runs_status)
    n_runs_failed      = n_runs - n_runs_succeeded
    pct_runs_succeeded = n_runs_succeeded / n_runs * 100

    comparsions_done      = filter(!ismissing, comparsion_success)
    n_compars             = length(comparsions_done)
    n_compars_succeeded   = count(comparsions_done)
    n_compars_failed      = n_compars - n_compars_succeeded
    pct_compars_succeeded = n_compars_succeeded / n_compars * 100

    final_result = (n_runs_succeeded == n_runs) && (n_compars_succeeded == n_compars)

    println(logs, repeat('=', n_header))
    println(logs, repeat('+', n_header))
    println(logs, repeat('=', n_header))
    println(logs, """

Summary

Tests succeded:         $(@sprintf("%3d/%-3d = %3.1f%%", n_runs_succeeded, n_runs, pct_runs_succeeded))
Comparsions succeded:   $(@sprintf("%3d/%-3d = %3.1f%%", n_compars_succeeded, n_compars, pct_compars_succeeded))
Final result:           $(final_result ? TERM_GREEN*"PASSED"*TERM_RESET : TERM_BLINK_RED*"FAILED"*TERM_RESET)
    """)

    logfilename = joinpath(TESTDIR, "tests.log")
    logstr = String(take!(logs))
    println(logstr)
    write(logfilename, logstr)

    errlogfilename = joinpath(TESTDIR, "errors.log")
    errlogstr = String(take!(errlogs))
    if length(errlogstr) > 0
        println("""Extended error messages were written to
                      $errlogfilename""")
        write(errlogfilename, errlogstr)
    end


    return final_result
end

runtests(tests::String...) = runtests(collect(tests))

function runtests(rgx::Regex)
    testconfig = load_testconfig()
    allreftests = collect(keys(testconfig))
    tests = filter(allreftests) do test
        contains(test, rgx)
    end
    if length(tests) == 0
        error("No tests found using regex $rgx. Make sure there are matching output folders in 'refs/' and a configuration entry in 'refs/testconfig.toml'")
    end
    return runtests(tests)
end

end
