#######################################################################
#                         Conservative fixing                         #
#######################################################################


@with_signature function conservative_fixing(equation::AbstractEquation)
  @accepts D, S, tau, x

  @unpack eos = equation
  atm_density = equation.atmosphere.rho
  atm_threshold = equation.atmosphere.threshold
  rhomin = atm_density * atm_threshold

  # B = 1.0 # in special relativity

  #####
  # conservative variable fixing following Deppe et al. arXiv:2109.12033
  ######

  fixed = false

  if D < rhomin || tau < 0.0
    D = rhomin
    tau = 1e-12
    fixed = true
  end

  # if D < 0.001 * rhomin
  #   # @warn "fixing D $D @ x = $x"
  #   D = rhomin
  #   fixed = true
  # end
  #
  # if tau < 0.0
  #   # @warn "fixing tau $tau @ x = $x"
  #   tau = 1e-12
  #   fixed = true
  # end
  # fix tau such that v < 0.999
  # if tau < 0.0
  #   @warn "fixing tau $tau @ x = $x"
  #   # What = 1 + tau / D = 1 / sqrt(1 - v^2)
  #   v2 = 0.999
  #   tau = max(D * (1.0 / sqrt(1.0 - v2) - 1), 1e-12)
  #   display(tau)
  #   fixed = true
  # end

  # (48)-(50) in arXiv:2109.12033 we put B=0, ̂μ = 0
  # solution to (52) is then
  What = 1.0 + tau / D

  S2 = S^2
  D2 = D^2
  ϵs = 1e-12
  factor = sqrt( (1.0-ϵs) * (What^2-1.0) * D2 / (S2 + D2 * 1e-16) )
  # @show factor, What
  if factor < 1.0 && What > 1.0
    # @warn "fixing ..."
    Sold = S
    S = 0.99 * sign(S) * sqrt( (1.0-ϵs) * (What^2-1.0) * D2 - D2 * 1e-16 )
    # @warn "S  = $Sold => $S @ x = $x"
    S2 = S^2
    factor = max(1.0, sqrt( (1.0-ϵs) * (What^2-1.0) * D2 / (S2 + D2 * 1e-16) ))
    # @show factor
    @assert factor >= 1.0
    fixed = true
  end

  # h0 = 1e-3
  # S2 = S^2 / B^2
  # r2 = S2 / D^2
  # z02 = r2 / h0^2
  # v02 = z02 / (1.0 + z02)
  # max_v2 = 0.999
  # if v02 > max_v2
  #   S = sign(S) * sqrt(max_v2/(1-max_v2) * h0^2 * D^2)
  # end

  flag_consfixed = Float64(fixed)

  @returns D, S, tau, flag_consfixed
end


#######################################################################
#                  cons2prim a la Kastaun et al 2021                  #
#######################################################################


@with_signature [simd=false] function cons2prim_kastaun(equation::AbstractEquation)
  @accepts D, S, tau, x

  @unpack eos = equation
  atm_density = equation.atmosphere.rho
  atm_threshold = equation.atmosphere.threshold
  rhomin = atm_density * atm_threshold
  rhomax = Inf
  rhoatm = atm_density
  # epsmin adjusted for IdealGas
  epsmin, epsmax = 0.0, Inf
  h0 = 1e-3

  B = 1.0 # in special relativity

  ####
  # cons2prim from Kastaun et al. arXiv:2005.01821
  ####

  q = tau / D
  S2 = (S/B)^2
  Si = S
  ri = Si / D
  # r = sqrt(S2) / D # =^= r in paper, r already denotes radius here
  z0i = ri / h0
  v0i = z0i / sqrt(1.0 + z0i^2)
  h02 = h0^2
  r2 = ri^2
  r = abs(ri)
  v0 = abs(v0i)

  if v0 > 1
    error((v0,z0i,Si,D))
  end

  # What is not a question but rather W^hat, cf. (40)
  fn_What = let r = r, v0 = v0
    mu -> begin
      vhat = min(mu * r, v0)
      return 1.0 / sqrt(1.0 - vhat^2)
    end
  end

  # setup bracket
  mu_a = 0.0
  # with no EM fields we can find the root of (49) analytically, because r = const.
  mu_plus = 1.0 / sqrt(h02 + r2)
  mu_b = r < h0 ? 1.0 / h0 : mu_plus

  What_a = fn_What(mu_a)
  What_b = fn_What(mu_b)
  if D / What_b > rhomax
    error("Invalid state encountered!")
  end

  if D < rhomin
    @warn "atmosphering I"

    error((D,S,tau))

    @assert eos isa EquationOfState.IdealGas
    @unpack Gamma = eos

    # employ atmosphere values by taking the zero temperature limit
    # for an ideal gas eos, we have that eos.Gamma becomes the Gamma for
    # a polytrope, see grhd notes for derivation

    rho = rhoatm
    v = 0.0
    W = 1.0
    # p = K * rho^Gamma
    # eps = Gamma * rho^
    # D = W * rho
    # S = rho * (1+eps+p/rho) * W^2 * v
    # tau = rho * (1+eps+p/rho) * W^2 - p - D

    # if eos isa ColdEquationOfState
    #   eps = eos.cold_eos(InternalEnergy, Density, rho)
    #   p = eos.cold_eos(Pressure, Density, rho)
    # else
    #   eps = eos(InternalEnergy, Density, rho)
    #   p = eos(Pressure, Density, rho)
    # end

  else

    # root bracket adjustment, cf. Appendix
    if D / What_b < rhomax < D
      @warn "adjusting upper root bracket"

      # infamous closure bug ...
      # https://docs.julialang.org/en/v1/manual/performance-tips/#man-performance-captured
      fn_rhomax = let D = D, rhomax = rhomax
        mu -> D / fn_What(mu) - rhomax
      end

      mu_b = find_zero(fn_rhomax, (mu_a, mu_plus), Roots.A42())

      if isnan(mu_b)
        error("Failed to correct upper interval bound")
      end

      println("New interval: ($mu_a, $mu_b)")
    end

    if D / What_b < rhomin < D # yes, we need What_b here
      @warn "adjusting lower root bracket"

      # infamous closure bug ...
      # https://docs.julialang.org/en/v1/manual/performance-tips/#man-performance-captured
      fn_rhomin = let D = D, rhomin = rhomin
        mu -> D / fn_What(mu) - rhomin
      end

      mu_a = find_zero(fn_rhomin, (mu_a, mu_b), Roots.A42())

      if isnan(mu_b)
        error("Failed to correct lower interval bound")
      end

      println("New interval: ($mu_a, $mu_b)")
    end

    # infamous closure bug ...
    # https://docs.julialang.org/en/v1/manual/performance-tips/#man-performance-captured
    master_fn = let D = D, v0 = v0, r = r, q = q, r2 = r2
      mu -> begin # eq (44)

        _vhat = min(mu * r, v0)
        _What = 1.0 / sqrt(1.0 - _vhat^2)
        rhohat = D / _What

        # eq (42)
        epshat = _What * (q - mu * r2) + _vhat^2 * _What^2 / (1.0 + _What)

        # TODO enforce EOS bounds - needed?

        # compute thermodynamic vars
        phat = eos(Pressure, Density, rhohat, InternalEnergy, epshat)
        # hhat = eos(Enthalpy, Density, rhohat, InternalEnergy, epshat)
        ahat = phat / (rhohat * (1.0 + epshat))

        # eq (46)-(48)
        nu_A = (1.0 + ahat) * (1.0 + epshat) / _What
        nu_B = (1.0 + ahat) * (1.0 + q - mu * r2)
        nu = max(nu_A, nu_B)

        # eq (44)
        f = mu - 1.0 / (nu + r2 * mu)
        return f
      end
    end

    # verify if there is a solution inside bracket
    mf_a = master_fn(mu_a)
    mf_b = master_fn(mu_b)
    mf_ab = mf_a*mf_b
    if mf_a * mf_b >= 0 || !isfinite(mf_ab)
      println((; D, S, tau, mf_a, mf_b, mu_a, mu_b, x))
      error("Master function is ill conditioned!")
    end

    mu = find_zero(master_fn, (mu_a, mu_b), Roots.A42() #= =^= algorithm 748 =#)
    if isnan(mu)
      error("Failed to find root")
    end

    #####
    # evaluate primitives
    #####

    # eq (36) [Kastaun] : h*W   = 1/mu
    # eq (15) [Kastaun] : rho*W = D
    rhohW2 = D / mu

    # regularize velocity according to Dumbser et al arxiv:2307.06629
    # eqs (82)-(84)
    y, y0 = rhohW2, 1e-4
    # if y > y0
      # v = S / mu
      # compute v accoriding to eq (68) or (40) [Kastaun]
      v = sign(Si) * min(mu * r, v0)
    # else
    #   # @warn "limiting v"
    #   # limit velocity ratio v/S
    #   fy = 2.0 * (y/y0)^3 - 3.0 * (y/y0)^2 + 1.0
    #   v = Si * y / (y^2 + fy * 5e-9)
    # end

    # eq (68) or (40)
    # v = min(mu * r, v0)
    W = 1.0 / sqrt(1.0 - v^2)
    # eq (41)
    rho = D / W
    # eq (42)
    eps = W * (q - mu * r2) + v^2 * W^2 / (1.0 + W)
    # eq (38)
    # eps2 = W * (q - mu * r2) + W - 1

    # enforce EOS bounds

    # Deppe at all also have a atmopshere treatment for this
    # if rhoatm <= rho <= rhomin
    #   @assert v^2 < 1e-4
    # end

    @assert rho > rhomin
    if eps < epsmin
      # adjusting tau such that energy density is within bounds
      # cf. 3rd paragraph in sec. III.A in Kastaun paper
      eps = epsmin
      taub4 = tau
      # invert this for tau: eps = W * (q - mu * r2) + v^2 * W^2 / (1.0 + W), q = tau / D
      tau = (eps / W  - v^2 * W / (1.0 + W) + mu * r2) * D
    end
    # if rho < rhomin || eps < epsmin
    #   @warn "atmosphering II @ x = $x : (rho,eps) = ($rho, $eps)"
    #   flag_cons2prim_atm2 = 1
    #   # atmosphering
    #   rho, eps, v, W = max(rhomin, D), epsmin, 0.0, 1.0
    #   D = rho
    #   tau =
    #   # rho, eps, v, W = rhomin, epsmin, 0.0, 1.0
    # elseif rho > rhomax || eps > epsmax
    #   @warn "rhomax || epsmax reached"
    #   flag_cons2prim_atm2 = 1
    #   rho, eps, v = rhomax, epsmax, Si*mu/D
    #   D = rho
    # end

    # compute thermodynamic vars
    p = eos(Pressure, Density, rho, InternalEnergy, eps)
    if p < 0
      error((rho,eps))
    end

  end

  # @returns D, S, tau, rho, v, eps, p
  @returns tau, rho, v, eps, p
end
