Project(env::Environment, prms) = Project(env, prms, env.mesh)


function Project(env::Environment, prms, mesh::Mesh1d)

  # construct Project
  eos = EquationOfState.make_EquationOfState(prms["EquationOfState"])
  equation = make_Equation(eos, prms["SRHD"], mesh)
  tci  = TCI.make_TCI(env.mesh, prms["TCI"])
  hrsc = HRSC.make_HRSC(env.mesh, prms["HRSC"])

  fixedprms = (; av_regularization=:covariant, id_smooth=true)
  P = Project(equation, hrsc, tci, fixedprms)

  # register variables
  # TODO Somehow replace _register_variables! with register_variables!
  @unpack cache = env
  _register_variables!(mesh, P)
  display(cache)

  # setup boundary conditions
  bdryconds = dg1d.DirichletBC2()
  ldg_bdryconds, av_bdryconds = nothing, nothing
  _register_variables!(mesh, bdryconds)

  # TODO Bdry conditions should be moved to project and should only be just a 'lightweight'
  # descriptor. Furthermore, they should also register variables in cache themselves.
  # Note: Atm, only the 'from_initial_data' option would require a place to store the data
  # somewhere. For periodic BCs we don't need extra storage at all.
  # But what other types of boundary conditions will be needed in the future?

  # setup initial data
  initialdata!(env, P, prms["SRHD"])

  # setup callbacks
  projectcb = make_callback(env, P, isperiodic(mesh))

  append!(env.callbacks, CallbackSet(projectcb.callbacks...))

  return P, (bdryconds, ldg_bdryconds, av_bdryconds)
end


function Project(env::Environment, prms, mesh::Mesh2d)

  # construct Project
  eos = EquationOfState.make_EquationOfState(prms["EquationOfState"])
  hrsc = HRSC.make_HRSC(mesh, prms["HRSC"])
  equation = make_Equation(eos, prms["SRHD"], mesh)
  bdrycond = dg1d.DirichletBC2()

  P = Project2d(equation, hrsc, bdrycond)

  # register variables
  # TODO add a ::Nothing overload for register_variables!
  # TODO Somehow replace _register_variables! with register_variables!
  @unpack cache = env
  _register_variables!(mesh, P)
  _register_variables!(mesh, hrsc, nothing)
  _register_variables!(mesh, bdrycond)
  display(cache)

  # TODO Bdry conditions should be moved to project and should only be just a 'lightweight'
  # descriptor. Furthermore, they should also register variables in cache themselves.
  # Note: Atm, only the 'from_initial_data' option would require a place to store the data
  # somewhere. For periodic BCs we don't need extra storage at all.
  # But what other types of boundary conditions will be needed in the future?

  # setup initial data
  initialdata!(env, P, prms["SRHD"])

  # setup boundary conditions
  bdryconds_x = nothing
  bdryconds_y = nothing
  ldg_bdryconds, av_bdryconds = nothing, nothing

  projectcb = make_callback(env, P)
  append!(env.callbacks, CallbackSet(projectcb.callbacks...))

  return P, ((bdryconds_x, bdryconds_y), ldg_bdryconds, av_bdryconds)
end


#######################################################################
#                         Register variables                          #
#######################################################################


# Julia limitation/bug: Overloading imported functions without their Module prefix shadows their
# definition in this module without a warning or error. This does not happen for methods
# defined in Base.
# Because of that we prefix with _ below.
# Cf. https://github.com/JuliaLang/julia/issues/15510
function _register_variables!(mesh::Mesh, P::Project)
  register_variables!(mesh.cache,
    dynamic_variablenames = (:D, :S, :tau,),            # conservatives
    rhs_variablenames     = (:rhs_D, :rhs_S, :rhs_tau), # rhs of conservatives
    static_variablenames  = (:flx_D, :flx_S, :flx_tau,  # conservatives' fluxes
                             :rho, :v, :p, :eps,        # primitives
                             :max_v,                    # maximum charactersitic speed
                             :E, :Em1,                  # entropy
                             :flx_E, :flx_Em1,          # entropy flux
                             :EP,                       # entropy production
                             :flag_consfixed),
    bdry_variablenames    = (:nflx_D, :nflx_S, :nflx_tau,
                             :bdry_D, :bdry_S, :bdry_tau, :bdry_max_v,
                             :bdry_rho, :bdry_v, :bdry_eps, :bdry_p,
                             :EPjump_lhs, :EPjump_rhs), # entropy production due to jumps at interfaces
    cell_variablenames    = (:cellmax_v,),
    global_variablenames  = (:t, :tm1) # time steps
  )
  _register_variables!(mesh, P.hrsc, P.prms)
  _register_variables!(mesh, P.tci)
  @unpack x = get_static_variables(mesh.cache)
  # @. x = env.mesh.x[:]
end


function _register_variables!(mesh, tci_or_hrsc::Nothing, args...) end


_register_variables!(mesh, hrsc::AbstractHRSC, prms) =
  error("_register_variables: Not implemented for hrsc of type '$(typeof(hrsc))'")

_register_variables!(mesh, hrsc::HRSC.AbstractReconstruction, prms) = nothing


function _register_variables!(mesh, hrsc::HRSC.AbstractArtificialViscosity, prms)
  register_variables!(mesh.cache,
    static_variablenames  = (:ldg_D, :ldg_S, :ldg_tau,
                             :flx_ldg_D, :flx_ldg_S, :flx_ldg_tau),
    cell_variablenames    = (:mu,:tmp_mu), # 'one viscosity to rule them all'
    )
end


function _register_variables!(mesh, hrsc::HRSC.SmoothedArtificialViscosity, prms)
  _register_variables!(mesh, hrsc.av, prms)
  register_variables!(mesh.cache,
    static_variablenames  = (:smoothed_mu,),
    bdry_variablenames    = (:bdry_ldg_D, :bdry_ldg_S, :bdry_ldg_tau,
                             :bdry_smoothed_mu),
  )
  reg = prms.av_regularization
  if reg == :covariant
    register_variables!(mesh.cache,
      bdry_variablenames    = (:bdry_rhs_D, :bdry_rhs_S, :bdry_rhs_tau)
    )
  end
end


function _register_variables!(mesh::Mesh2d, hrsc::HRSC.EntropyViscosity, prms)
  register_variables!(mesh.cache,
    static_variablenames = (:E, :Em1, :flx_E_x, :flx_E_y, :flxm1_E_x, :flxm1_E_y,
                            :ldg_D_x, :ldg_D_y, :ldg_Sx_x, :ldg_Sx_y,
                            :ldg_Sy_x,  :ldg_Sy_y,  :ldg_tau_x,  :ldg_tau_y),
    bdry_variablenames   = (:nflx_D_x, :nflx_D_y, :nflx_Sx_x, :nflx_Sx_y,
                            :nflx_Sy_x, :nflx_Sy_y, :nflx_tau_x, :nflx_tau_y),
    cell_variablenames   = (:mu,),
    global_variablenames = (:tm1,)
  )
end


function _register_variables!(mesh, tci::TCI.AbstractTCI)
  register_variables!(mesh.cache, cell_variablenames = (:flag,:D_flag,:S_flag,:tau_flag))
end


function _register_variables!(mesh, tci::TCI.EntropyProduction)
  register_variables!(mesh.cache,
                      cell_variablenames = (:flag,),
  )
end


function _register_variables!(mesh, P::Project2d)
  register_variables!(mesh.cache,
    dynamic_variablenames = (:D, :Sx, :Sy, :tau,),            # conservatives
    rhs_variablenames     = (:rhs_D, :rhs_Sx, :rhs_Sy, :rhs_tau), # rhs of conservatives
    static_variablenames  = (:flx_D_x, :flx_D_y,
                             :flx_Sx_x, :flx_Sx_y,
                             :flx_Sy_x, :flx_Sy_y,
                             :flx_tau_x, :flx_tau_y,          # conservatives' fluxes
                             :rho, :vx, :vy, :p, :eps,        # primitives
                             :max_v),                         # maximum charactersitic speed
    bdry_variablenames    = (:nflx_D, :nflx_Sx, :nflx_Sy, :nflx_tau),
    global_variablenames  = (:t,),
    cell_variablenames    = (:cellmax_v,),
  )
end


function _register_variables!(mesh::Mesh1d, bdrycond::dg1d.DirichletBC2)
  register_variables!(mesh.cache,
    static_variablenames  = (:init_D, :init_S, :init_tau, :init_p, :init_v, :init_max_v)
  )
end


function _register_variables!(mesh::Mesh2d, bdrycond::dg1d.DirichletBC2)
  register_variables!(mesh.cache,
    static_variablenames  = (:init_D, :init_Sx, :init_Sy, :init_tau,
                             :init_flx_D_x, :init_flx_D_y,
                             :init_flx_Sx_x, :init_flx_Sx_y,
                             :init_flx_Sy_x, :init_flx_Sy_y,
                             :init_flx_tau_x, :init_flx_tau_y,
                             :init_max_v)
  )
end
