Project(env::Environment, prms) = Project(env, env.mesh, prms)


function Project(env::Environment, mesh::Mesh1d, prms)

  # construct Project
  equation = make_Equation(prms["ScalarEq"], dimension(mesh))
  tci  = TCI.make_TCI(env.mesh, prms["TCI"])
  hrsc = HRSC.make_HRSC(env.mesh, prms["HRSC"])

  # TODO remove this
  bdryconds, ldg_bdryconds, av_bdryconds = nothing, nothing, nothing

  bernstein = BernsteinReconstruction(mesh)
  av_derivative_scheme = Symbol(prms["ScalarEq"]["av_derivative_scheme"])
  av_drag = prms["ScalarEq"]["av_drag"]
  fixedprms = (; bernstein, av_derivative_scheme, av_drag)
  P = Project(equation, hrsc, tci, bdryconds, fixedprms)

  # register variables
  # TODO Somehow replace _register_variables! with register_variables!
  @unpack cache = env
  _register_variables!(mesh, equation)
  _register_variables!(mesh, tci)
  _register_variables!(mesh, hrsc)
  display(cache)

  # setup initial data
  initialdata!(env, P, prms["ScalarEq"])

  # TODO Bdry conditions should be moved to project and should only be just a 'lightweight'
  # descriptor. Furthermore, they should also register variables in cache themselves.
  # Note: Atm, only the 'from_initial_data' option would require a place to store the data
  # somewhere. For periodic BCs we don't need extra storage at all.
  # But what other types of boundary conditions will be needed in the future?

  # setup callbacks
  projectcb = make_callback(env, P, isperiodic(mesh))
  append!(env.callbacks, CallbackSet(projectcb.callbacks...))

  return P, (bdryconds, ldg_bdryconds, av_bdryconds)
end


function Project(env::Environment, mesh::Mesh2d, prms)

  # construct Project
  equation = make_Equation(prms["ScalarEq"], dimension(mesh))
  tci  = TCI.make_TCI(mesh, prms["TCI"])
  hrsc = HRSC.make_HRSC(mesh, prms["HRSC"])

  bdrycond = dg1d.DirichletBC2()
  fixedprms = (;)
  P = Project(equation, hrsc, tci, bdrycond, fixedprms)

  # register variables
  # TODO Somehow replace _register_variables! with register_variables!
  @unpack cache = env
  _register_variables!(mesh, equation)
  _register_variables!(mesh, tci)
  _register_variables!(mesh, hrsc)
  _register_variables!(cache, bdrycond)
  display(cache)

  # setup initial data
  initialdata!(env, P, prms["ScalarEq"])

  # # setup boundary conditions
  bdryconds_x = nothing
  bdryconds_y = nothing
  ldg_bdryconds, av_bdryconds = nothing, nothing

  # setup callbacks
  projectcb = make_callback(env, P, isperiodic(mesh))
  append!(env.callbacks, CallbackSet(projectcb.callbacks...))

  return P, ((bdryconds_x, bdryconds_y), ldg_bdryconds, av_bdryconds)
end


#######################################################################
#                         Register variables                          #
#######################################################################


function _register_variables!(cache, tci) end


# Julia limitation/bug: Overloading imported functions without their Module prefix shadows their
# definition in this module without a warning or error. This does not happen for methods
# defined in Base.
# Because of that we prefix with _ below.
# Cf. https://github.com/JuliaLang/julia/issues/15510
# Perhaps this is the import vs. using confusion people talk on the forums.


function _register_variables!(mesh::Mesh1d, eq::AbstractScalarEq)
  register_variables!(mesh.cache,
    dynamic_variablenames = (:u,),
    rhs_variablenames     = (:rhs_u,),
    static_variablenames  = (:flx_u, :src_u, :E, :Em1, :F, :Fm1, :v),
    cell_variablenames    = (:max_v,),
    bdry_variablenames    = (:nflx_u,:bdry_u),
    global_variablenames  = (:t, :tm1))
end
function _register_variables!(mesh::Mesh2d, eq::AbstractScalarEq)
  register_variables!(mesh.cache,
    dynamic_variablenames = (:u,),
    rhs_variablenames     = (:rhs_u,),
    static_variablenames  = (:flx_x, :src_u, :flx_y, :v, :v_x, :v_y),
    cell_variablenames    = (:max_v,),
    bdry_variablenames    = (:nflx_u, :bdry_u),
    global_variablenames  = (:t, :tm1))
end


function _register_variables!(mesh::Mesh1d, hrsc::HRSC.AbstractArtificialViscosity)
  register_variables!(mesh.cache,
    static_variablenames  = (:ldg_u,),
    bdry_variablenames    = (:bdry_ldg_u, :bdry_smoothed_mu),
    cell_variablenames    = (:mu,:mu_m1,))
end

function _register_variables!(mesh::Mesh2d, hrsc::HRSC.AbstractArtificialViscosity)
  register_variables!(mesh.cache,
    static_variablenames  = (:ldg_u_x, :ldg_u_y),
    bdry_variablenames    = (:bdry_ldg_u_x, :bdry_ldg_u_y, :bdry_smoothed_mu,
                             :nflx_u_x, :nflx_u_y),
    cell_variablenames    = (:mu,))
end
function _register_variables!(mesh::Mesh2d, hrsc::HRSC.EntropyViscosity)
  register_variables!(mesh.cache,
    static_variablenames  = (:ldg_u_x, :ldg_u_y, :E, :Em1, :F_x, :F_y, :F_xm1, :F_ym1),
    bdry_variablenames    = (:bdry_ldg_u_x, :bdry_ldg_u_y, :bdry_smoothed_mu,
                             :nflx_u_x, :nflx_u_y),
    cell_variablenames    = (:mu,))
end


function _register_variables!(mesh::Mesh1d, hrsc::HRSC.SmoothedArtificialViscosity)
  _register_variables!(mesh, hrsc.av)
  register_variables!(mesh.cache, static_variablenames  = (:smoothed_mu,))
end
function _register_variables!(mesh::Mesh2d, hrsc::HRSC.SmoothedArtificialViscosity)
  _register_variables!(mesh, hrsc.av)
  register_variables!(mesh.cache, static_variablenames  = (:smoothed_mu,))
end


function _register_variables!(mesh, hrsc::Maybe{HRSC.AbstractHRSC}) end


function _register_variables!(mesh, tci::TCI.AbstractTCI)
  register_variables!(mesh.cache, cell_variablenames = (:flag,))
end


function _register_variables!(mesh::Mesh1d, tci::TCI.EntropyProduction)
  register_variables!(mesh.cache,
                      cell_variablenames   = (:flag,),
                      static_variablenames = (:EP,),
                      bdry_variablenames   = (:EPjump_lhs, :EPjump_rhs))
end


function _register_variables!(cache, bdryconds::Nothing) end
function _register_variables!(cache, bdryconds::dg1d.AbstractBC)
  register_variables!(cache,
    static_variablenames  = (:init_u, :init_flx_x, :init_flx_y, :init_v)
  )
end
