@testset "Callbacks" begin


@testset "CallbackTiming" begin

  ### disable periodic triggering, so any istimed! call will yield false
  cbt = CallbackTiming(every_iteration = 0, every_dt = 0, every_dt_walltime = 0)
  @test dg1d.istimed!(cbt, 0) == false
  @test dg1d.istimed!(cbt, 1) == false
  @test dg1d.istimed!(cbt, 2) == false
  @test dg1d.istimed!(cbt, 3) == false


  ### only trigger wrt iteration count
  cbt = CallbackTiming(every_iteration = 2)
  @test dg1d.istimed!(cbt, 0) == false
  @test dg1d.istimed!(cbt, 1) == true
  @test dg1d.istimed!(cbt, 2) == false
  @test dg1d.istimed!(cbt, 3) == true
  # calling with times that already passed don't increment iteration count
  @test dg1d.istimed!(cbt, 0.5) == false
  @test dg1d.istimed!(cbt, 1.5) == false
  @test dg1d.istimed!(cbt, 2.5) == false
  @test dg1d.istimed!(cbt, 5) == false # here we increment again
  @test dg1d.istimed!(cbt, 6) == true # here we trigger


  ### only trigger wrt time interval
  cbt = CallbackTiming(every_dt = 2)
  @test dg1d.istimed!(cbt, 0) == false
  @test dg1d.istimed!(cbt, 1) == false
  @test dg1d.istimed!(cbt, 2) == true
  @test dg1d.istimed!(cbt, 3) == false
  @test dg1d.istimed!(cbt, 4) == true

  ### only trigger wrt walltime time interval
  cbt = CallbackTiming(every_dt_walltime = 1.0)
  @test dg1d.istimed!(cbt, 1) == false
  sleep(1.0)
  @test dg1d.istimed!(cbt, 2) == true

  ### one can also mix every_iteration, every_dt
  cbt = CallbackTiming(every_iteration = 3, every_dt = 2)
  @test dg1d.istimed!(cbt, 0) == false
  @test dg1d.istimed!(cbt, 1) == false # triggered by every_iteration
  @test dg1d.istimed!(cbt, 2) == true  # triggered by every_dt and every_iteration
  @test dg1d.istimed!(cbt, 3) == false
  @test dg1d.istimed!(cbt, 4) == true  # triggered by every_dt
  @test dg1d.istimed!(cbt, 5) == true  # triggered by every_iteration

  # test constructor
  @test_throws AssertionError CallbackTiming(every_iteration = -1, every_dt = 0)
  @test_throws AssertionError CallbackTiming(every_iteration = 0, every_dt = -1)

end


@testset "FunctionCallback" begin

  counter = 0
  function increment_counter(u, t)
    counter += 1
    return true
  end

  fcb = FunctionCallback(increment_counter, CallbackTiming(every_iteration=1))
  # increment three times through callback
  fcb(0, 0, 0)
  fcb(0, 1, 0)
  fcb(0, 2, 0)
  @test counter == 3

  # force callback evaluation
  counter = 0
  fcb2 = FunctionCallback(increment_counter, CallbackTiming(every_iteration=100))
  fcb(0, 0, 0)
  fcb(0, 1, 0)
  fcb(0, 2, 0)
  @test counter == 0


end


@testset "CallbackSet" begin

  function increment(u, t)
    u .+= 1
    return true
  end

  fcb1 = FunctionCallback(increment, CallbackTiming(every_iteration=1,every_dt=0))
  fcb2 = FunctionCallback(increment, CallbackTiming(every_iteration=1,every_dt=0))

  cbs = CallbackSet(fcb1, fcb2)

  state_u = [0]

  # increment state_u six times through calling `increment` twice on each CallbackSet call
  # note: ts must increase
  cbs(state_u, 0)
  cbs(state_u, 1)
  cbs(state_u, 2)

  @test state_u[1] == 6

end


@testset "SaveCallback" begin

  @testset "1d Output" begin

    mktempdir() do dir

      mesh = Mesh1d()
      cache = dg1d.cache(mesh)
      dg1d.register_variables!(cache, static_variablenames=[:u, :v], cell_variablenames=[:a,:b])

      @unpack u, v = get_static_variables(cache)
      @unpack a, b = get_cell_variables(cache)
      u .= randn(length(mesh))
      v .= randn(length(mesh))
      a .= randn(n_cells(mesh))
      b .= randn(n_cells(mesh))

      filename = joinpath(dir, "testoutput1d")
      scb = SaveCallback([:u, :a], cache, mesh, filename,
                         CallbackTiming(every_iteration=1, every_dt=0),
                         format1d="h5")

      # save two times with time steps t=0, t=1
      t1, t2 = 0.0, 1.0
      scb(0, t1, 0)
      scb(0, t2, 0)

      @test isfile(filename)

      xx, us, as, ts = h5open(filename) do file
        groups = keys(file)
        filter!(g->g!="mesh", groups)
        groups = parse.(Int64, groups)
        sort!(groups)
        us, as = Vector{Float64}[], Vector{Float64}[]
        ts = Float64[]
        for g in groups
          push!(us, read(file[string(g)]["u"]))
          push!(as, read(file[string(g)]["a"]))
          push!(ts, read(file[string(g)]["t"]))
        end
        x = read(file["mesh"]["x"])
        return x, us, as, ts
      end

      @unpack x = get_static_variables(mesh.cache)
      @test all(isapprox(xx, x))
      for (ui, ai) in zip(us, as)
        @test all(isapprox(a, ai))
        @test all(isapprox(u, ui))
      end
      @test all(isapprox.(ts, [t1, t2]))

    end

  end

  @testset "2d Output" begin

    mktempdir() do dir

      mesh = Mesh2d()
      cache = dg1d.cache(mesh)
      dg1d.register_variables!(cache, static_variablenames=[:u, :v], cell_variablenames=[:a,:b])

      @unpack u, v = get_static_variables(cache)
      @unpack a, b = get_cell_variables(cache)
      u .= randn(length(mesh))
      v .= randn(length(mesh))
      a .= randn(n_cells(mesh))
      b .= randn(n_cells(mesh))

      #### HDF5 version
      filename = joinpath(dir, "testoutput2d_h5")
      scb = SaveCallback([:u, :a], cache, mesh, filename,
                         CallbackTiming(every_iteration=1, every_dt=0),
                         format2d="h5")

      # save two times with time steps t=0, t=1
      t1, t2 = 0.0, 1.0
      scb(0, t1, 0)
      scb(0, t2, 0)

      @test isfile(filename)

      x, y, us, as, ts = h5open(filename) do file
        groups = keys(file)
        filter!(g->g!="mesh", groups)
        groups = parse.(Int64, groups)
        sort!(groups)
        us, as = Vector{Float64}[], Vector{Float64}[]
        ts = Float64[]
        for g in groups
          push!(us, read(file[string(g)]["u"]))
          push!(as, read(file[string(g)]["a"]))
          push!(ts, read(file[string(g)]["t"]))
        end
        x = read(file["mesh"]["x"])
        y = read(file["mesh"]["y"])
        return x, y, us, as, ts
      end

      @test all(isapprox(x, get_static_variables(mesh.cache).x))
      @test all(isapprox(y, get_static_variables(mesh.cache).y))
      for (ui, ai) in zip(us, as)
        @test all(isapprox(a, ai))
        @test all(isapprox(u, ui))
      end
      @test all(isapprox.(ts, [t1, t2]))


      # 2d outputs

      #### XDMF version
      subdir = joinpath(dir, "xdmf")
      mkdir(subdir)
      filename = joinpath(subdir, "testoutput2d_xdmf")
      scb = SaveCallback([:u, :v], cache, mesh, filename,
                         CallbackTiming(every_iteration=1, every_dt=0),
                         format2d="xdmf" #= default is xdmf+h5 =#)
      # save two times with time steps t=0, t=1
      t1, t2 = 0.0, 1.0
      scb(0, t1, 0)
      scb(0, t2, 0)

      # too lazy to verify xdmf itself, just make sure something has been written
      outputdir = joinpath(dirname(filename), "xdmf")
      @test isdir(outputdir)
      @test isfile(joinpath(outputdir, "u.xdmf"))
      @test isfile(joinpath(outputdir, "v.xdmf"))
      @test isfile(joinpath(outputdir, "mesh.xdmf"))
      @test isfile(joinpath(outputdir, "data.h5"))

      #### HDF5+XDMF version
      subdir = joinpath(dir, "xdmf+h5")
      mkdir(subdir)
      filename = joinpath(subdir, "testoutput2d_xdmf+h5")
      scb = SaveCallback([:u, :v], cache, mesh, filename,
                         CallbackTiming(every_iteration=1, every_dt=0),
                         format2d="xdmf+h5" #= same as above =#)
      # save two times with time steps t=0, t=1
      t1, t2 = 0.0, 1.0
      scb(0, t1, 0)
      scb(0, t2, 0)

      # too lazy to verify xdmf itself, just make sure something has been written
      outputdir = joinpath(dirname(filename), "xdmf")
      @test isdir(outputdir)
      @test isfile(joinpath(outputdir, "u.xdmf"))
      @test isfile(joinpath(outputdir, "v.xdmf"))
      @test isfile(joinpath(outputdir, "mesh.xdmf"))
      @test isfile(joinpath(outputdir, "data.h5"))

      #### BIN+XDMF version
      # not yet implemented
      subdir = joinpath(dir, "xdmf+bin")
      mkdir(subdir)
      filename = joinpath(subdir, "testoutput2d_xdmf+bin")
      @test_broken scb = SaveCallback([:u, :v], cache, mesh, filename,
                         CallbackTiming(every_iteration=1, every_dt=0),
                         format2d="xdmf+bin" #= same as above =#)
      # # save two times with time steps t=0, t=1
      # t1, t2 = 0.0, 1.0
      # scb(0, t1, 0)
      # scb(0, t2, 0)
      #
      # # too lazy to verify xdmf itself, just make sure something has been written
      # outputdir = joinpath(dirname(filename), "xdmf")
      # @test isdir(outputdir)
      # @test isfile(joinpath(outputdir, "u.xdmf"))
      # @test isfile(joinpath(outputdir, "v.xdmf"))
      # @test isfile(joinpath(outputdir, "mesh.xdmf"))
      # @test isfile(joinpath(outputdir, "data.h5"))

      #### VTK version
      subdir = joinpath(dir, "vtk")
      mkdir(subdir)
      filename = joinpath(subdir, "testoutput2d_vtk")
      scb = SaveCallback([:u, :v], cache, mesh, filename,
                         CallbackTiming(every_iteration=1, every_dt=0),
                         format2d="vtk")
      # save two times with time steps t=0, t=1
      t1, t2 = 0.0, 1.0
      scb(0, t1, 0)
      scb(0, t2, 0)

      # too lazy to verify xdmf itself, just make sure something has been written
      outputdir = joinpath(dirname(filename), "vtk")
      @test isdir(outputdir)
      @test isfile(joinpath(outputdir, "data1.vtu"))
      @test isfile(joinpath(outputdir, "data2.vtu"))
      @test isfile(joinpath(outputdir, "timeseries.pvd"))

    end

  end

end


@testset "InterpolationCallback" begin

  @testset "1d Output" begin

    mktempdir() do dir

      mesh = Mesh1d(N=25,K=25)
      dg1d.register_variables!(mesh.cache,
                               static_variablenames=[:u, :v])

      @unpack x, u, v = get_static_variables(mesh.cache)
      @. u = sinpi(x)
      v .= randn(length(mesh))

      (xmin, xmax), = extrema(mesh)
      intrp_x = collect(range(xmin, xmax, length=100))
      intrp_u = @. sinpi(intrp_x)

      filename = joinpath(dir, "interpolate.h5")
      icb = InterpolationCallback([:u], mesh, intrp_x, filename,
                                  CallbackTiming(every_iteration=1, every_dt=0))

      # save two times with time steps t=0, t=1
      t1, t2 = 0.0, 1.0
      icb(0, t1, 0)
      icb(0, t2, 0)

      @test isfile(filename)

      x, us, ts = h5open(filename) do file
        groups = keys(file)
        filter!(g->g!="mesh", groups)
        groups = parse.(Int64, groups)
        sort!(groups)
        us = Vector{Float64}[]
        ts = Float64[]
        for g in groups
          push!(us, read(file[string(g)]["u"]))
          push!(ts, read(file[string(g)]["t"]))
        end
        x = read(file["mesh"]["x"])
        x, us, ts
      end

      @test all(isapprox(x, intrp_x))
      for ui in us
        @test all(isapprox(ui, intrp_u))
      end
      @test all(isapprox.(ts, [t1, t2]))

    end

    # can only interpolate variables that are registered as dynamic, static or rhs.
    mktempdir() do dir
      mesh = Mesh1d(N=25,K=25)
      dg1d.register_variables!(mesh.cache,
                               cell_variablenames=[:u, :v])
      (xmin, xmax), = extrema(mesh)
      intrp_x = collect(range(xmin, xmax, length=100))
      filename = joinpath(dir, "interpolate.h5")
      @test_throws "InterpolationCallback: Cannot interpolate variable 'u', because it is not registered as one of" begin
        InterpolationCallback([:u], mesh, intrp_x, filename,
                              CallbackTiming(every_iteration=1, every_dt=0))
      end
    end

  end

end


@testset "TimeAlignedSaveCallback" begin

  mktempdir() do dir

    mesh = Mesh1d()
    cache = dg1d.cache(mesh)
    dg1d.register_variables!(cache, dynamic_variablenames=[:u])

    # fixed time step
    timestep_fn(u, t, n) = 1.0

    # artificial rhs method
    u0 = zeros(length(mesh)) # initial data
    rhs!(du, u, t) = u .= randn(length(mesh))

    # save every dt=2
    filename = joinpath(dir, "testoutput.h5")
    scb = SaveCallback([:u], cache, mesh, filename, CallbackTiming(every_dt=2.0),
                       format1d="h5")

    # request time aligned save callback
    save_times = [ 0.5, 1.3, 3.1 ]
    atscb = TimeAlignedSaveCallback(save_times, timestep_fn, scb)
    aligned_timestep_fn = atscb.aligned_timestep_fn

    cbs = CallbackSet(scb, atscb)

    # evolve
    pblm = dg1d.ODEProblem(:rk4)
    tend = 5.0
    dg1d.evolve(rhs!, deepcopy(u0), aligned_timestep_fn, tend, dg1d.algorithm(:rk4),
                callback_fullstep=cbs)

    # test output
    @test isfile(filename)

    ts = h5open(filename) do file
      groups = keys(file)
      filter!(g->g!="mesh", groups)
      groups = parse.(Int64, groups)
      sort!(groups)
      ts = Float64[]
      for g in groups
        push!(ts, read(file[string(g)]["t"]))
      end
      return ts
    end

    # explanation for test results below:
    # we scheduled a SaveCallback with every_dt=2.0; if we evolve till t=5.0 then we should
    # see at least 2 outputs when starting from t=0.0
    # furthermore, we scheduled a TimeAlignedSaveCallback with three time stamps before t = 5.0,
    # hence, we expect 3 outputs for this one
    # note that we no longer force the callback evaluation at t=0.0, this needs to be done manually
    # in total we should find 5 outputs
    @test length(ts) == 5
    @test all(save_times .∈ Ref(ts))

    # test for wrong argument type of timestep_fn in constructor
    @test_throws AssertionError TimeAlignedSaveCallback(save_times, (u,t)->0.1, scb)
    @test_throws AssertionError TimeAlignedSaveCallback(save_times, (u,t::Vector,n)->0.1, scb)

  end

end


end
