@testset "Output" begin

  mktempdir() do dir

    mesh = Mesh1d(; N=4,K=6)
    cache = dg1d.cache(mesh)
    dg1d.register_variables!(cache, static_variablenames=[:u], cell_variablenames=[:a])

    filename = joinpath(dir, "testoutput.h5")
    scb = SaveCallback([:u, :a], cache, mesh, filename,
                       CallbackTiming(every_iteration=1, every_dt=0),
                       format1d="h5")

    # save some randomly generated data
    n_times = 5
    t = 0.0
    dt = 0.1 # artificial time spacing
    random_data = Dict(:u => [ randn(length(mesh)) for _ = 1:n_times ],
                       :a => [ randn(n_cells(mesh)) for _ = 1:n_times ])
    times = [ dt * (t_idx - 1) for t_idx = 1:n_times ]
    @unpack u = get_static_variables(cache)
    @unpack a = get_cell_variables(cache)
    for (t_idx, t) in enumerate(times)
      u .= random_data[:u][t_idx]
      a .= random_data[:a][t_idx]
      scb(nothing #= enough because we mutated arrays above =#, t, t_idx)
    end

    @test isfile(filename)

    # now test Output
    output = Output(filename)

    xx = output.x
    @unpack x = get_static_variables(mesh.cache)
    @test all(isapprox(xx, x))

    # test standard read
    # Note: SaveCallback stores all data as vectors.
    for (t_idx, t) in enumerate(times)
      ref_u = random_data[:u][t_idx]
      ref_a = random_data[:a][t_idx]
      # request data at time t
      @test all(isapprox(ref_u, read_variable(output, "u", t)))
      @test all(isapprox(ref_a, read_variable(output, "a", t)))
      # request data at iteration t_idx
      @test all(isapprox(ref_u, read_variable(output, "u", t_idx)))
      @test all(isapprox(ref_a, read_variable(output, "a", t_idx)))
    end

    # test reading into storage
    Npts, K = layout(mesh)
    storage_mat_u = zeros(Npts, K)
    storage_vec_u = zeros(Npts*K)
    storage_mat_u_wrong_shape = zeros(K, Npts)
    storage_vec_u_wrong_length = zeros(2*Npts*K)

    for (t_idx, _) in enumerate(times)
      ref_u = random_data[:u][t_idx]
      ref_a = random_data[:a][t_idx]
      # read into a vector
      read_variable!(storage_vec_u, output, "u", t_idx)
      @test all(isapprox(ref_u, storage_vec_u))
      # read into a matrix
      read_variable!(storage_mat_u, output, "u", t_idx)
      @test all(isapprox(ref_u, storage_mat_u[:]))
      # attempt reads into wrongly sized storages
      @test_throws ErrorException read_variable!(storage_vec_u_wrong_length, output,
                                              "u", t_idx)
      # Attention: Reading into wrongly shaped matrix does not throw!
      read_variable!(storage_mat_u_wrong_shape, output, "u", t_idx)
    end

  end

end
