function make_data(N, nin, nout)
  args = Tuple( rand(N) for _ = 1:nin )
  rets = Tuple( rand(N) for _ = 1:nout )
  return args, rets, 1:N, 1:N
end


@testset "@with_signature" begin

  N = 100

  struct DummyEquation end
  dummyeq = DummyEquation()
  struct DummyMesh end
  mesh = DummyMesh()

  # general usage
  @with_signature function fn_general(eq::DummyEquation)
    @accepts u1, u2
    ru1, ru2 = 2.0 * u1 + u2, u2 * 3.0 - u1
    @returns ru1, ru2
  end
  @test accepts(fn_general, dummyeq, mesh) == (:u1, :u2)
  @test returns(fn_general, dummyeq, mesh) == (:ru1, :ru2)
  @test has_signature(fn_general, dummyeq, mesh) == true
  args, rets, vidxs, bidxs = make_data(N, 2, 2)
  fn_general(dummyeq, mesh, args, rets, vidxs, bidxs)
  u1, u2 = args
  ru1, ru2 = deepcopy(rets)
  @. ru1 = 2.0 * u1 + u2
  @. ru2 = u2 * 3.0 - u1
  expected = (ru1, ru2)
  @test rets == expected

  # general usage dispatch on two arguments
  @with_signature function fn_general2(eq::DummyEquation, mesh::DummyMesh)
    @accepts u1, u2
    ru1, ru2 = 2.0 * u1 + u2, u2 * 3.0 - u1
    @returns ru1, ru2
  end
  @test accepts(fn_general, dummyeq, mesh) == (:u1, :u2)
  @test returns(fn_general, dummyeq, mesh) == (:ru1, :ru2)
  @test has_signature(fn_general, dummyeq, mesh) == true
  args, rets, vidxs, bidxs = make_data(N, 2, 2)
  fn_general(dummyeq, mesh, args, rets, vidxs, bidxs)
  u1, u2 = args
  ru1, ru2 = deepcopy(rets)
  @. ru1 = 2.0 * u1 + u2
  @. ru2 = u2 * 3.0 - u1
  expected = (ru1, ru2)
  @test rets == expected

  # @accepts and @returns with only one argument each
  @with_signature function fn_accepts1(eq::DummyEquation)
    @accepts u1
    ru1 = 2.0 * u1
    @returns ru1
  end
  @test accepts(fn_accepts1, dummyeq, mesh) == (:u1,)
  @test returns(fn_accepts1, dummyeq, mesh) == (:ru1,)
  @test has_signature(fn_accepts1, dummyeq, mesh) == true
  args, rets, vidxs, bidxs = make_data(N, 1, 1)
  fn_accepts1(dummyeq, mesh, args, rets, vidxs, bidxs)
  u1, = args
  ru1, = deepcopy(rets)
  @. ru1 = 2.0 * u1
  expected = (ru1,)
  @test rets == expected

  # multiple @accepts statements
  @with_signature function fn_accepts2(eq::DummyEquation)
    @accepts u1
    @accepts v1, v2
    ru1 = 2.0 * u1
    rv1 = -3.0 * v1
    rv2 = -3.0 * v2
    @returns ru1, rv1, rv2
  end
  @test accepts(fn_accepts2, dummyeq, mesh) == (:u1,:v1,:v2)
  @test returns(fn_accepts2, dummyeq, mesh) == (:ru1,:rv1,:rv2)
  @test has_signature(fn_accepts2, dummyeq, mesh) == true
  args, rets, vidxs, bidxs = make_data(N, 3, 3)
  fn_accepts2(dummyeq, mesh, args, rets, vidxs, bidxs)
  u1, v1, v2 = args
  ru1, rv1, rv2 = deepcopy(rets)
  @. ru1 = 2.0 * u1
  @. rv1 = -3.0 * v1
  @. rv2 = -3.0 * v2
  expected = (ru1, rv1, rv2)
  @test rets == expected

  # Prefix @accepts variables
  @with_signature function fn_prefix1(eq::DummyEquation)
    @accepts [Prefix(tmp)] u1
    ru1 = 2.0 * tmp_u1
    @returns ru1
  end
  @test accepts(fn_prefix1, dummyeq, mesh) == (:u1,)
  @test returns(fn_prefix1, dummyeq, mesh) == (:ru1,)
  @test has_signature(fn_prefix1, dummyeq, mesh) == true
  args, rets, vidxs, bidxs = make_data(N, 1, 1)
  fn_prefix1(dummyeq, mesh, args, rets, vidxs, bidxs)
  u1,  = args
  ru1, = deepcopy(rets)
  @. ru1 = 2.0 * u1
  expected = (ru1,)
  @test rets == expected

  # Prefix multiple @accepts variables
  @with_signature function fn_prefix2(eq::DummyEquation)
    @accepts [Prefix(lhs)] u1, u2
    @accepts [Prefix(rhs)] u1, u2
    @accepts n1, n2
    flx_u1 = n1 * lhs_u1 + n2 * rhs_u1
    flx_u2 = n1 * lhs_u2 + n2 * rhs_u2
    @returns flx_u1, flx_u2
  end
  @test accepts(fn_prefix2, dummyeq, mesh) == (:u1,:u2,:u1,:u2,:n1,:n2)
  @test returns(fn_prefix2, dummyeq, mesh) == (:flx_u1,:flx_u2)
  @test has_signature(fn_prefix2, dummyeq, mesh) == true
  args, rets, vidxs, bidxs = make_data(N, 6, 2)
  fn_prefix2(dummyeq, mesh, args, rets, vidxs, bidxs)
  lhs_u1, lhs_u2, rhs_u1, rhs_u2, n1, n2 = args
  flx_u1, flx_u2 = deepcopy(rets)
  @. flx_u1 = n1 * lhs_u1 + n2 * rhs_u1
  @. flx_u2 = n1 * lhs_u2 + n2 * rhs_u2
  expected = (flx_u1, flx_u2)
  @test rets == expected

  # no variable name for dispatch type required
  @with_signature function fn_noargname(#= missing name here =#::DummyEquation)
    @accepts u1
    ru1 = 2.0 * u1
    @returns ru1
  end
  @test accepts(fn_noargname, dummyeq, mesh) == (:u1,)
  @test returns(fn_noargname, dummyeq, mesh) == (:ru1,)
  @test has_signature(fn_noargname, dummyeq, mesh) == true
  args, rets, vidxs, bidxs = make_data(N, 1, 1)
  fn_noargname(dummyeq, mesh, args, rets, vidxs, bidxs)
  u1, = args
  ru1, = deepcopy(rets)
  @. ru1 = 2.0 * u1
  expected = (ru1,)
  @test rets == expected

  # using only bdry variables
  @with_signature function fn_bdry1(eq::DummyEquation)
    @accepts [bdry] u1, u2
    @accepts [bdry] n1, n2
    flx_u1 = n1 * u1 + n2 * u1
    flx_u2 = n1 * u2 + n2 * u2
    @returns [bdry] flx_u1, flx_u2
  end
  @test accepts(fn_bdry1, dummyeq, mesh) == (:u1,:u2,:n1,:n2)
  @test returns(fn_bdry1, dummyeq, mesh) == (:flx_u1,:flx_u2)
  @test has_signature(fn_bdry1, dummyeq, mesh) == true
  args, rets, vidxs, bidxs = make_data(N, 4, 2)
  fn_bdry1(dummyeq, mesh, args, rets, vidxs, bidxs)
  u1, u2, n1, n2 = args
  flx_u1, flx_u2 = deepcopy(rets)
  @. flx_u1 = n1 * u1 + n2 * u1
  @. flx_u2 = n1 * u2 + n2 * u2
  expected = (flx_u1,flx_u2)
  @test rets == expected

  # using volume and bdry variables with prefixes
  @with_signature function fn_bdry2(eq::DummyEquation)
    @accepts u1
    @accepts [Prefix(test)] u2
    @accepts [Prefix(foo), bdry] n1, n2
    bar_flx_u1 = foo_n1 * u1 + foo_n2 * u1
    foo_flx_u2 = foo_n1 * test_u2 + foo_n2 * test_u2
    @returns [Prefix(bar)] flx_u1
    @returns [bdry,Prefix(foo)] flx_u2
  end
  @test accepts(fn_bdry2, dummyeq, mesh) == (:u1,:u2,:n1,:n2)
  @test returns(fn_bdry2, dummyeq, mesh) == (:flx_u1,:flx_u2)
  @test has_signature(fn_bdry2, dummyeq, mesh) == true
  args, rets, vidxs, bidxs = make_data(N, 4, 2)
  fn_bdry1(dummyeq, mesh, args, rets, vidxs, bidxs)
  u1, u2, n1, n2 = args
  flx_u1, flx_u2 = deepcopy(rets)
  @. flx_u1 = n1 * u1 + n2 * u1
  @. flx_u2 = n1 * u2 + n2 * u2
  expected = (flx_u1,flx_u2)
  @test rets == expected

  # using only bdry variables, but with shuffled indices
  # simple version first: compute @. flx_u1[rng_bidxs] = u1[rng_bidxs], which
  # does not alter the ordering of values, but instead just computes in different order
  @with_signature function fn_bdry3(eq::DummyEquation)
    @accepts [bdry] u1
    flx_u1 = u1
    @returns [bdry] flx_u1
  end
  @test accepts(fn_bdry3, dummyeq, mesh) == (:u1,)
  @test returns(fn_bdry3, dummyeq, mesh) == (:flx_u1,)
  @test has_signature(fn_bdry3, dummyeq, mesh) == true
  args, rets, vidxs, bidxs = make_data(10, 1, 1)
  rng_bidxs = Random.shuffle(bidxs)
  fn_bdry3(dummyeq, mesh, args, rets, vidxs, rng_bidxs)
  u1, = args
  flx_u1, = deepcopy(rets)
  @. flx_u1 = u1
  expected = (flx_u1,)
  @test rets == expected

  # using only bdry variables, but with shuffled indices
  @with_signature function fn_bdry4(eq::DummyEquation)
    @accepts [bdry] u1, u2
    @accepts [bdry] n1, n2
    flx_u1 = n1 * u1 + n2 * u1
    flx_u2 = n1 * u2 + n2 * u2
    @returns [bdry] flx_u1, flx_u2
  end
  @test accepts(fn_bdry4, dummyeq, mesh) == (:u1,:u2,:n1,:n2)
  @test returns(fn_bdry4, dummyeq, mesh) == (:flx_u1,:flx_u2)
  @test has_signature(fn_bdry4, dummyeq, mesh) == true
  args, rets, vidxs, bidxs = make_data(N, 4, 2)
  rng_bidxs = Random.shuffle(bidxs)
  fn_bdry4(dummyeq, mesh, args, rets, vidxs, rng_bidxs)
  u1, u2, n1, n2 = args
  flx_u1, flx_u2 = deepcopy(rets)
  @. flx_u1 = n1 * u1 + n2 * u1
  @. flx_u2 = n1 * u2 + n2 * u2
  expected = (flx_u1,flx_u2)
  @test rets == expected

  # use volume and bdry variables with shuffled bdry indices
  @with_signature function fn_volbdry1(eq::DummyEquation)
    @accepts u1, u2
    @accepts [bdry] n1, n2
    flx_u1 = n1 * u1 + n2 * u1
    flx_u2 = n1 * u2 + n2 * u2
    @returns [bdry] flx_u1, flx_u2
  end
  @test accepts(fn_volbdry1, dummyeq, mesh) == (:u1,:u2,:n1,:n2)
  @test returns(fn_volbdry1, dummyeq, mesh) == (:flx_u1,:flx_u2)
  @test has_signature(fn_volbdry1, dummyeq, mesh) == true
  args, rets, vidxs, bidxs = make_data(N, 4, 2)
  rng_bidxs = Random.shuffle(bidxs)
  fn_volbdry1(dummyeq, mesh, args, rets, vidxs, rng_bidxs)
  u1, u2, n1, n2 = args
  flx_u1, flx_u2 = deepcopy(rets)
  @. flx_u1[rng_bidxs] = n1[rng_bidxs] * u1 + n2[rng_bidxs] * u1
  @. flx_u2[rng_bidxs] = n1[rng_bidxs] * u2 + n2[rng_bidxs] * u2
  expected = (flx_u1,flx_u2)
  @test rets == expected

  # use volume and bdry variables with shuffled volume indices
  @with_signature function fn_volbdry2(eq::DummyEquation)
    @accepts u1, u2
    @accepts [bdry] n1, n2
    flx_u1 = n1 * u1 + n2 * u1
    flx_u2 = n1 * u2 + n2 * u2
    @returns [bdry] flx_u1, flx_u2
  end
  @test accepts(fn_volbdry2, dummyeq, mesh) == (:u1,:u2,:n1,:n2)
  @test returns(fn_volbdry2, dummyeq, mesh) == (:flx_u1,:flx_u2)
  @test has_signature(fn_volbdry2, dummyeq, mesh) == true
  args, rets, vidxs, bidxs = make_data(N, 4, 2)
  rng_vidxs = Random.shuffle(vidxs)
  fn_volbdry2(dummyeq, mesh, args, rets, rng_vidxs, bidxs)
  u1, u2, n1, n2 = args
  flx_u1, flx_u2 = deepcopy(rets)
  @. flx_u1 = n1 * u1[rng_vidxs] + n2 * u1[rng_vidxs]
  @. flx_u2 = n1 * u2[rng_vidxs] + n2 * u2[rng_vidxs]
  expected = (flx_u1,flx_u2)
  @test rets == expected

  # use volume and bdry variables with shuffled volume and bdry indices
  @with_signature function fn_volbdry3(eq::DummyEquation)
    @accepts u1, u2
    @accepts [bdry] n1, n2
    flx_u1 = n1 * u1 + n2 * u1
    flx_u2 = n1 * u2 + n2 * u2
    @returns [bdry] flx_u1, flx_u2
  end
  @test accepts(fn_volbdry3, dummyeq, mesh) == (:u1,:u2,:n1,:n2)
  @test returns(fn_volbdry3, dummyeq, mesh) == (:flx_u1,:flx_u2)
  @test has_signature(fn_volbdry3, dummyeq, mesh) == true
  args, rets, vidxs, bidxs = make_data(N, 4, 2)
  rng_vidxs = Random.shuffle(vidxs)
  rng_bidxs = Random.shuffle(bidxs)
  fn_volbdry3(dummyeq, mesh, args, rets, rng_vidxs, rng_bidxs)
  u1, u2, n1, n2 = args
  flx_u1, flx_u2 = deepcopy(rets)
  @. flx_u1[rng_bidxs] = n1[rng_bidxs] * u1[rng_vidxs] + n2[rng_bidxs] * u1[rng_vidxs]
  @. flx_u2[rng_bidxs] = n1[rng_bidxs] * u2[rng_vidxs] + n2[rng_bidxs] * u2[rng_vidxs]
  expected = (flx_u1,flx_u2)
  @test rets == expected

  # use longer volume and bdry variables than (shuffled) volume and bdry index ranges
  @with_signature function fn_volbdry4(eq::DummyEquation)
    @accepts u1, u2
    @accepts [bdry] n1, n2
    flx_u1 = n1 * u1 + n2 * u1
    flx_u2 = n1 * u2 + n2 * u2
    @returns [bdry] flx_u1, flx_u2
  end
  @test accepts(fn_volbdry3, dummyeq, mesh) == (:u1,:u2,:n1,:n2)
  @test returns(fn_volbdry3, dummyeq, mesh) == (:flx_u1,:flx_u2)
  @test has_signature(fn_volbdry3, dummyeq, mesh) == true
  args, rets, _, _ = make_data(N, 4, 2)
  rng_vidxs = Random.shuffle(1:div(N,2))
  rng_bidxs = Random.shuffle(rng_vidxs) .+ 10 # reshuffled and offset
  fn_volbdry3(dummyeq, mesh, args, rets, rng_vidxs, rng_bidxs)
  u1, u2, n1, n2 = args
  flx_u1, flx_u2 = deepcopy(rets)
  @. flx_u1[rng_bidxs] = n1[rng_bidxs] * u1[rng_vidxs] + n2[rng_bidxs] * u1[rng_vidxs]
  @. flx_u2[rng_bidxs] = n1[rng_bidxs] * u2[rng_vidxs] + n2[rng_bidxs] * u2[rng_vidxs]
  expected = (flx_u1[rng_bidxs],flx_u2[rng_bidxs])
  obtained = (rets[1][rng_bidxs],rets[2][rng_bidxs])
  @test obtained == expected


  # test options
  # debug=true disables all 'dangerous' optimizations like @inbounds, @simd
  @with_signature [debug=true] function fn_debug1(eq::DummyEquation)
    @accepts u1
    ru1 = 2.0 * u1 + 1.0
    @returns ru1
  end
  @test accepts(fn_debug1, dummyeq, mesh) == (:u1,)
  @test returns(fn_debug1, dummyeq, mesh) == (:ru1,)
  @test has_signature(fn_debug1, dummyeq, mesh) == true
  args, rets, vidxs, bidxs = make_data(N, 1, 1)
  fn_debug1(dummyeq, mesh, args, rets, vidxs, bidxs)
  (u1,), (ru1,) = args, deepcopy(rets)
  u1, = args
  ru1, = deepcopy(rets)
  @. ru1 = 2.0 * u1 + 1.0
  expected = (ru1,)
  @test rets == expected

  # test options
  @with_signature [simd=false] function fn_debug2(eq::DummyEquation)
    @accepts u1
    ru1 = 2.0 * u1 + 1.0
    @returns ru1
  end
  @test accepts(fn_debug2, dummyeq, mesh) == (:u1,)
  @test returns(fn_debug2, dummyeq, mesh) == (:ru1,)
  @test has_signature(fn_debug2, dummyeq, mesh) == true
  args, rets, vidxs, bidxs = make_data(N, 1, 1)
  fn_debug2(dummyeq, mesh, args, rets, vidxs, bidxs)
  (u1,), (ru1,) = args, deepcopy(rets)
  u1, = args
  ru1, = deepcopy(rets)
  @. ru1 = 2.0 * u1 + 1.0
  expected = (ru1,)
  @test rets == expected


  # test options
  @with_signature [inbounds=false] function fn_debug3(eq::DummyEquation)
    @accepts u1
    ru1 = 2.0 * u1 + 1.0
    @returns ru1
  end
  @test accepts(fn_debug3, dummyeq, mesh) == (:u1,)
  @test returns(fn_debug3, dummyeq, mesh) == (:ru1,)
  @test has_signature(fn_debug3, dummyeq, mesh) == true
  args, rets, vidxs, bidxs = make_data(N, 1, 1)
  fn_debug3(dummyeq, mesh, args, rets, vidxs, bidxs)
  (u1,), (ru1,) = args, deepcopy(rets)
  u1, = args
  ru1, = deepcopy(rets)
  @. ru1 = 2.0 * u1 + 1.0
  expected = (ru1,)
  @test rets == expected


  ##### error handling

  # error on invalid options syntax
  @test_throws LoadError @eval @with_signature { inbounds=false } function fn_undefvar(eq::DummyEquation)
    @accepts u1
    ru1 = 2.0 * u1
    @returns ru1, undef_var
  end
  @test_throws LoadError @eval @with_signature inbounds=true function fn_undefvar(eq::DummyEquation)
    @accepts u1
    ru1 = 2.0 * u1
    @returns ru1, undef_var
  end

  # error on duplicated @with_signature options
  @test_throws LoadError @eval @with_signature [inbounds=true,inbounds=false] function fn_undefvar(eq::DummyEquation)
    @accepts u1
    ru1 = 2.0 * u1
    @returns ru1, undef_var
  end
  @test_throws LoadError @eval @with_signature [skip_debug_info=true,inbounds=false,inbounds=false] function fn_undefvar(eq::DummyEquation)
    @accepts u1
    ru1 = 2.0 * u1
    @returns ru1, undef_var
  end
  # error on invalid type for option
  @test_throws LoadError @eval @with_signature [inbounds=1] function fn_undefvar(eq::DummyEquation)
    @accepts u1
    ru1 = 2.0 * u1
    @returns ru1, undef_var
  end
  @test_throws LoadError @eval @with_signature [unknown_option=true,inbounds=false] function fn_unknown_option(eq::DummyEquation)
    @accepts u1
    ru1 = 2.0 * u1
    @returns ru1
  end

  # function with unused variable ru2 in body throws normal UndefVarError
  @with_signature function fn_undefvar(eq::DummyEquation)
    @accepts u1
    ru1 = 2.0 * u1
    @returns ru1, undef_var
  end
  args, rets, vidxs, bidxs = make_data(N, 1, 2)
  @test_throws UndefVarError fn_undefvar(dummyeq, mesh, args, rets, vidxs, bidxs)

  # missing function body error
  @test_throws LoadError @eval @with_signature function fn(eq::DummyEquation) end

  # @accepts/@returns must appear only at the start/end of the function body
  @test_throws LoadError @eval @with_signature function fn_w_error(eq::DummyEquation)
    println("this line is not allowed here")
    @accepts u1
    ru1 = 2.0 * u1
    @returns ru1
  end
  @test_throws LoadError @eval @with_signature function fn_w_error(eq::DummyEquation)
    @accepts u1
    ru1 = 2.0 * u1
    @returns ru1
    println("this line is not allowed")
  end
  @test_throws LoadError @eval @with_signature function fn_w_error(eq::DummyEquation)
    println("this line is not allowed here")
    @accepts u1
    @accepts u2
    ru1 = 2.0 * u1
    @returns ru1
  end
  @test_throws LoadError @eval @with_signature function fn_w_error(eq::DummyEquation)
    @accepts u1
    ru1 = 2.0 * u1
    @returns ru1
    println("this line is not allowed")
    @returns ru2
  end
  @test_throws LoadError @eval @with_signature function fn_w_error(eq::DummyEquation)
    @accepts u1
    ru1 = 2.0 * u1
    @returns ru1
    ru1 = 3.0 * u1
    @returns ru2
  end

  # @accepts and @returns require each at least one argument
  @test_throws LoadError @eval @with_signature function fn_w_error(eq::DummyEquation)
    @accepts
    ru1 = 1
    @returns ru1
  end
  @test_throws LoadError @eval @with_signature function fn_w_error(eq::DummyEquation)
    @accepts u1
    ru1 = 2.0 * u1
    @returns
  end
  @test_throws LoadError @eval @with_signature function fn_w_error(eq::DummyEquation)
    @accepts
    @accepts u1
    ru1 = 1
    @returns ru1
  end
  @test_throws LoadError @eval @with_signature function fn_w_error(eq::DummyEquation)
    @accepts u1
    ru1 = 2.0 * u1
    @returns ru1
    @returns
  end

  # error on duplicated options for @accepts/@returns
  @test_throws LoadError @eval @with_signature function fn_w_error(eq::DummyEquation)
    @accepts [bdry,bdry] u1
    ru1 = 2.0 * u1
    @returns ru1
  end
  @test_throws LoadError @eval @with_signature function fn_w_error(eq::DummyEquation)
    @accepts [Prefix(lhs),bdry,Prefix(lhs)] u1
    ru1 = 2.0 * u1
    @returns ru1
  end
  @test_throws LoadError @eval @with_signature function fn_w_error(eq::DummyEquation)
    @accepts [blabla] u1
    ru1 = 2.0 * u1
    @returns ru1
  end

  # function definition requires one or two arguments
  @test_throws LoadError @eval @with_signature function fn_w_error(#=no argument=#)
    @accepts u1
    ru1 = 2.0 * u1
    @returns ru1
  end
  @test_throws LoadError @eval @with_signature function fn_w_error(eq::DummyEquation, mesh, too_much)
    @accepts u1
    ru1 = 2.0 * u1
    @returns ru1
  end

  # the only argument must be type annotated
  @test_throws LoadError @eval @with_signature function fn_w_error(eq)
    @accepts u1
    ru1 = 2.0 * u1
    @returns ru1
  end

  # fallback interface methods error when called
  function fn_with_no_signature(eq::DummyEquation)
    @accepts u1
    ru1 = 2.0 * u1
    @returns ru1
  end
  @test_throws ErrorException accepts(fn_with_no_signature,dummyeq,mesh)
  @test_throws ErrorException returns(fn_with_no_signature,dummyeq,mesh)
  # except the trait function is just false
  has_signature(fn_with_no_signature, dummyeq, mesh) == false

end
